/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.io.FileInputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Properties;

import org.apache.log4j.PropertyConfigurator;
import org.bbaw.pdr.padre.PadreConfiguration;
import org.bbaw.pdr.padre.client.PadreClientException;
import org.bbaw.pdr.solr.SolrConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Fabian Körner
 */
public final class Configuration
{

	// FIXME keep version number uptodate
	private static final String PDR_ALLIES_VERSION_NUMBER = "1.1.0";
	private static final String PDR_ALLIES_SHORT_INFO = "PDR Allies v" + PDR_ALLIES_VERSION_NUMBER;

	private static final Configuration INSTANCE = new Configuration();

	private static boolean _isLoggingInitialised = false;
	private static Logger _logger;

	private static Properties _properties;

	private Configuration()
	{
		Configuration.intialiseLogging();
		Configuration.configurePDRAllies();
	}

	/**
	 * Singleton: get unique instance.
	 * @return Unique instance.
	 */
	public static Configuration getInstance()
	{
		return INSTANCE;
	}

	private static void intialiseLogging()
	{
		if (!_isLoggingInitialised)
		{
			String log4jConfigurationFilename = Configuration.buildLog4jConfigurationFilename();

			PropertyConfigurator.configure(log4jConfigurationFilename);
			_logger = LoggerFactory.getLogger("org.bbaw.pdr.allies");

			_logger.info("logging for {} initialised.", PDR_ALLIES_SHORT_INFO);
		}
	}

	private static String buildLog4jConfigurationFilename()
	{
		StringBuilder sb = new StringBuilder();
		String pdrHome = System.getenv("PDR_HOME");
		String fileSeparator = System.getProperty("file.separator");

		sb.append(pdrHome);
		sb.append(fileSeparator);
		sb.append("conf");
		sb.append(fileSeparator);
		sb.append("log4j.conf");
		return sb.toString();
	}

	private static String buildPDRConfigurationFilename()
	{
		StringBuilder sb = new StringBuilder();
		String pdrHome = System.getenv("PDR_HOME");
		String fileSeparator = System.getProperty("file.separator");

		sb.append(pdrHome);
		sb.append(fileSeparator);
		sb.append("conf");
		sb.append(fileSeparator);
		sb.append("pdr.conf");
		_logger.debug("configuration filename: " + sb.toString());
		return sb.toString();
	}

	private static void configurePDRAllies()
	{
		_properties = Configuration.loadPDRAlliesProperties();

		URI padreBaseURI = Configuration.getURIFromString(_properties.getProperty("PADRE_BASE_URI"));
		String padreUsername = _properties.getProperty("PADRE_USERNAME");
		String padrePassword = _properties.getProperty("PADRE_PASSWORD");

		_logger.debug("padre credentials: " + padreBaseURI.toString() + ", " + padreUsername + ", " + ((padrePassword == null) ? "null" : "*****"));
		try
		{
			PadreConfiguration.getInstance().setCredentials(padreBaseURI, padreUsername, padrePassword);
		}
		catch (PadreClientException cause)
		{
			_logger.error("PADRE CONFIGURATION FAILED!", cause);
		}

		String solrBase = _properties.getProperty("SOLR_BASE_URI");
		_logger.debug("solr base: " + solrBase);
		URI solrBaseURI = Configuration.getURIFromString(solrBase);

		SolrConfiguration.getInstance().setBaseURI(solrBaseURI);
	}

	private static URI getURIFromString(final String uri)
	{
		URI result = null;
		try
		{
			result = new URI(uri);
		}
		catch (URISyntaxException use)
		{
			_logger.error("invalid URI: " + uri, use);
		}
		return result;
	}

	private static Properties loadPDRAlliesProperties()
	{
		Properties result = new Properties();
		try
		{
			String configurationFilename = Configuration.buildPDRConfigurationFilename();
			FileInputStream fis = new FileInputStream(configurationFilename);
			result.load(fis);
			fis.close();
		}
		catch (IOException ioe)
		{
			_logger.error("couldn't load pdr configuration", ioe);
		}
		return result;
	}

	/**
	 * Returns system property <code>catalina.home</code>.
	 * @return Catalina home.
	 */
	public String getCatalinaHome()
	{
		return System.getProperty("catalina.home");
	}

	/**
	 * Returns PDR-Allies version number.
	 * @return Version number.
	 */
	public String getVersionNumber()
	{
		return PDR_ALLIES_VERSION_NUMBER;
	}

	/**
	 * Returns PDR-Allies short info.
	 * @return Short info.
	 */
	public String getShortInfo()
	{
		return PDR_ALLIES_SHORT_INFO;
	}

	/**
	 * Sends provided messages to logging system at DEBUG level.
	 * @param strings Messages.
	 */
	public void logDebug(final String... strings)
	{
		if (_logger.isDebugEnabled())
		{
			StringBuilder sb = new StringBuilder();
			for (String string : strings)
			{
				sb.append(string);
			}
			_logger.debug(sb.toString());
		}
	}

	/**
	 * Sends provided message to logging system at ERROR level.
	 * @param message Error message.
	 */
	public void logError(final String message)
	{
		_logger.error(message);
	}

	/**
	 * Sends provided error information to logging system.
	 * @param message Error message.
	 * @param throwable Further detail.
	 */
	public void logError(final String message, final Throwable throwable)
	{
		_logger.error(message, throwable);
	}
}
