/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.bbaw.pdr.allies.error.InvalidCredentialsException;

/**
 * Authentication credentials for transmission to server-side.
 * @author Fabian Körner
 */
public class Credentials
{

	private static final Pattern PATTERN_STRING_FORM = Pattern.compile("^pdrUo[.](\\d{3})[.](\\d{3})[.](\\d{9})[#][0-9a-f]{32}$");

	private String _credentials;
	private String _userID;
	private String _md5Password;

	/**
	 * Constructor with combined single <code>java.lang.String</code>. Parts must be separated by '<code>#</code>'.
	 * @param credentials Combined single string.
	 * @throws InvalidCredentialsException Provided credentials failed to pass format validation.
	 */
	public Credentials(final String credentials) throws InvalidCredentialsException
	{
		_credentials = credentials;
		Credentials.validate(_credentials);
		String[] parts = Credentials.getParts(credentials);
		_userID = parts[0];
		_md5Password = parts[1];
	}

	/**
	 * Constructor with separated parts.
	 * @param userID User ID.
	 * @param md5Password MD5 hash of password.
	 * @throws InvalidCredentialsException Provided credentials failed to pass format validation.
	 */
	public Credentials(final String userID, final String md5Password) throws InvalidCredentialsException
	{
		_credentials = userID + "#" + md5Password;
		Credentials.validate(_credentials);
		_userID = userID;
		_md5Password = md5Password;
	}

	/**
	 * Returns combined <code>java.lang.String</code> of user ID and password hash.
	 * @return Combination of user ID and password hash.
	 */
	public final String getCredentials()
	{
		return _credentials;
	}

	/**
	 * Returns user's ID.
	 * @return User ID.
	 */
	public final String getUserID()
	{
		return _userID;
	}

	/**
	 * Return MD5 hash of user's password.
	 * @return MD5 hash of password.
	 */
	public final String getMD5Password()
	{
		return _md5Password;
	}

	private static String[] getParts(final String credentials) throws InvalidCredentialsException
	{
		String[] parts = credentials.split("#");
		if (2 == parts.length)
		{
			return parts;
		}
		else
		{
			throw new InvalidCredentialsException("invalid credentials [" + credentials + "]");
		}
	}

	private static void validate(final String credentials) throws InvalidCredentialsException
	{
		Matcher matcher = PATTERN_STRING_FORM.matcher(credentials);
		if (!matcher.find() || 0 == Integer.parseInt(matcher.group(1)) || 0 == Integer.parseInt(matcher.group(2)) || 0 == Integer.parseInt(matcher.group(3)))
		{
			throw new InvalidCredentialsException("invalid credentials [" + credentials + "]");
		}
	}
}
