/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.util.HashMap;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.bbaw.pdr.padre.PadreSingleReadXml;
import org.bbaw.pdr.solr.SolrHttpQuery;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRObject;

/**
 * Only for use with PIT.
 * @author Fabian Körner
 */
public class PITConnector
{
	private static final Configuration CONFIGURATION = Configuration.getInstance();

	/**
	 * Create a Solr query from provided parameters and returns the unchanged
	 * response.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with Solr's response.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement solrSelect(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
		CONFIGURATION.logDebug("input #solrSelect: ", inElement.toString());

    	final HashMap<String, String> parameters = new HashMap<String, String>();
    	OMElement currentChildElement;
    	String key;
    	String value;
    	for (Iterator<?> i = inElement.getChildElements(); i.hasNext();)
    	{
    		currentChildElement = (OMElement) i.next();
    		key = currentChildElement.getLocalName();
    		value = Tools.uncloak(currentChildElement.getText());
        	CONFIGURATION.logDebug("parameter parsed #solrSelect: [" + key + " = " + value + "]");
    		parameters.put(key, value);
    	}

    	String response = null;
    	try
    	{
        	response = new SolrHttpQuery(parameters).execute();
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("PITConnector#solrSelect: ", e);
			throw new AxisFault(e.getMessage() + " (PITConnector#solrSelect)");
    	}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_SOLR_SELECT_RESPONSE);
    	OMElement responseElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_RESPONSE);
    	responseElement.setText(Tools.cloak(response));
    	outElement.addChild(responseElement);
		CONFIGURATION.logDebug("output #solrSelect: ", outElement.toString());

    	return outElement;
	}

	/**
	 * Returns the PDR object of the provided ID.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with PDR object.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getObject(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getObject: ", inElement.toString());

		OMElement idElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_ID));

		/* Ergebnis */
		PDRObject pdrObject;
		try
		{
	    	Identifier id = new Identifier(idElement.getText());
			pdrObject = new PadreSingleReadXml(id.toString()).execute();
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("PITConnector#getObject: ", e);
			throw new AxisFault(e.getMessage() + " (PITConnector#getObject)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_OBJECT_RESPONSE);
		OMElement objectElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECT);
		objectElement.setText(Tools.cloak(pdrObject.getXMLContent()));
		outElement.addChild(objectElement);
		CONFIGURATION.logDebug("output #getObject: ", outElement.toString());

		return outElement;
	}
}
