/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.bbaw.pdr.utils.Identifier;

/**
 * Encapsulates the access to the PDR repository.
 * @author Fabian Körner
 */
public class Repository
{
	// TODO join classes Utilities and Repository or rethink methods of both
	private static final Configuration CONFIGURATION = Configuration.getInstance();

	/**
	 * Returns objects which were modified since a given timestamp.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with objects which were modified since provided timestamp.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getModifiedObjects(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
		CONFIGURATION.logDebug("input #getModifiedObjects: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

    	OMElement instanceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_INSTANCE));
    	OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));
    	OMElement timestampElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_TIMESTAMP));
    	int instance = Integer.parseInt(instanceElement.getText());
    	int project = Integer.parseInt(projectElement.getText());
    	String timestamp = timestampElement.getText();

    	/* Ergebnis */
    	List<String> objects = new Vector<String>();
    	try
    	{
        	objects = org.bbaw.pdr.utils.Repository.getModifiedObjects(instance, project, timestamp);
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("Repository#getModifiedObjects: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#getModifiedObjects)");
    	}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_GET_MODIFIED_OBJECTS_RESPONSE);
    	OMElement objectsElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECTS);
    	OMElement objectElement;
    	for (String object : objects)
    	{
    		objectElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECT);
    		objectElement.setText(Tools.cloak(object));
    		objectsElement.addChild(objectElement);
    	}
    	outElement.addChild(objectsElement);
    	CONFIGURATION.logDebug("output #getModifiedObjects: ", outElement.toString());

    	return outElement;
	}

	/**
	 * Returns object counts of all availabe projects.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with sizes of all available objects.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getProjectSizes(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getProjectSizes: ", inElement.toString());

    	/* Ergebnis */
		Map <Integer, Integer> sizesMap = new HashMap<Integer, Integer>();
		try
		{
			sizesMap = org.bbaw.pdr.utils.Repository.getProjectSizes();
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Repository#getProjectSizes: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#getProjectSizes)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_GET_PROJECT_SIZES_RESPONSE);
		OMElement mapElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_MAP);
		for (Integer key : sizesMap.keySet())
		{
			OMElement pairElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_KEY_VALUE_PAIR);
			OMElement keyElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_KEY);
			OMElement valueElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_VALUE);

			keyElement.setText(Tools.cloak(key.toString()));
			valueElement.setText(Tools.cloak(sizesMap.get(key).toString()));

			pairElement.addChild(keyElement);
			pairElement.addChild(valueElement);
			mapElement.addChild(pairElement);
		}
		outElement.addChild(mapElement);
		CONFIGURATION.logDebug("output #getProjectSizes: ", outElement.toString());

		return outElement;
	}

	/**
	 * Translates a given project ID and username into the user's PDR ID.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with user ID.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getUserID(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
    	CONFIGURATION.logDebug("input #getUserID: ", inElement.toString());

		OMElement usernameElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_USER));
		OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));

		String username = usernameElement.getText();
		int project = Integer.parseInt(projectElement.getText());

    	/* Ergebnis */
    	Identifier identifier;
    	try
    	{
    		identifier = org.bbaw.pdr.utils.Repository.getUserID(username, project);
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("Repository#getUserID: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#getUserID)");
    	}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_GET_USER_ID_REPSONSE);
    	OMElement idElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_ID);
    	idElement.setText(identifier.toString());
    	outElement.addChild(idElement);
    	CONFIGURATION.logDebug("output #getUserID: ", outElement.toString());

    	return outElement;
	}

	/**
	 * Returns the current server time.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with server timestamp.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getTime(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
    	CONFIGURATION.logDebug("input #getTime: ", inElement.toString());

    	/* Ergebnis */
    	String timestamp;
    	try
    	{
    		timestamp = org.bbaw.pdr.utils.Repository.getTime();
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("Repository#getTime: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#getTime)");
    	}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_GET_TIME_REPSONSE);
    	OMElement idElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_TIMESTAMP);
    	idElement.setText(timestamp);
    	outElement.addChild(idElement);
    	CONFIGURATION.logDebug("output #getTime: ", outElement.toString());

    	return outElement;
	}

	/**
	 * Ingests given PDR objects, and returns a mapping of temporary to persistent IDs.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with a map of temporary and persistent IDs.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement ingestObjects(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
		CONFIGURATION.logDebug("input #ingestObjects: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

    	OMElement instanceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_INSTANCE));
    	OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));

    	int instance = Integer.parseInt(instanceElement.getText());
    	int project = Integer.parseInt(projectElement.getText());
    	Vector<String> objects = new Vector<String>();
    	int count = 0;
    	OMElement currentObjectElement;
    	String object;
    	while ((currentObjectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_OBJECT
    			+ Integer.toString(count++)))) != null)
    	{
    		object = Tools.uncloak(currentObjectElement.getText());
    		objects.add(object);
    	}

    	/* Ergebnis */
    	Map<Identifier, Identifier> identifierMap = new HashMap<Identifier, Identifier>();
    	try
    	{
    		identifierMap = org.bbaw.pdr.utils.Repository.ingestObjects(instance, project, objects);
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("Repository#ingestObjects: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#ingestObjects)");
    	}

	    /* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(
    			AlliesConstants.OMCHILD_REPOSITORY_INGEST_OBJECTS_RESPONSE);
    	OMElement mapElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_MAP);
    	for (Identifier key : identifierMap.keySet())
    	{
    		OMElement pairElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_KEY_VALUE_PAIR);
    		OMElement keyElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_KEY);
    		OMElement valueElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_VALUE);

    		keyElement.setText(Tools.cloak(key.toString()));
    		valueElement.setText(Tools.cloak(identifierMap.get(key).toString()));

    		pairElement.addChild(keyElement);
    		pairElement.addChild(valueElement);
    		mapElement.addChild(pairElement);
    	}
    	outElement.addChild(mapElement);
		CONFIGURATION.logDebug("output #ingestObjects: ", outElement.toString());

    	return outElement;
	}

	/**
	 * Writes object modifications into the repository. If conflicts occur, the server-side
	 * versions of conflicting objects will be returned.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with conflicting objects.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement modifyObjects(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
		CONFIGURATION.logDebug("input #modifyObjects: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

    	OMElement instanceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_INSTANCE));
    	OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));
    	OMElement forceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_FORCE));
    	int instance = Integer.parseInt(instanceElement.getText());
    	int project = Integer.parseInt(projectElement.getText());
    	boolean force = Boolean.parseBoolean(forceElement.getText());
    	Vector<String> modifiedObjects = new Vector<String>();
    	int count = 0;
    	OMElement currentObjectElement;
    	String objectString;
    	while ((currentObjectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_OBJECT + Integer.toString(count++)))) != null)
    	{
    		objectString = Tools.uncloak(currentObjectElement.getText());
    		modifiedObjects.add(objectString);
    	}

    	/* Ergebnis */
    	List<String> conflictingObjects = new Vector<String>();
    	try
    	{
    		conflictingObjects = org.bbaw.pdr.utils.Repository.modifyObjects(instance, project, modifiedObjects, force);
    	}
    	catch (Exception e)
    	{
			CONFIGURATION.logError("Repository#modifyObjects: ", e);
			throw new AxisFault(e.getMessage() + " (Repository#modifyObjects)");
    	}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_REPOSITORY_MODIFY_OBJECTS_RESPONSE);
    	OMElement objectsElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECTS);
    	OMElement objectElement;
    	for (String object : conflictingObjects)
    	{
    		objectElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECT);
    		objectElement.setText(Tools.cloak(object));
    		objectsElement.addChild(objectElement);
    	}
    	outElement.addChild(objectsElement);
		CONFIGURATION.logDebug("output #modifyObjects: ", outElement.toString());

    	return outElement;
	}
}
