/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.SSLContext;
import javax.net.ssl.X509TrustManager;
import javax.xml.namespace.QName;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.util.AXIOMUtil;
import org.apache.axis2.AxisFault;
import org.apache.http.HttpResponse;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.bbaw.pdr.allies.error.PDRAlliesClientException;
import org.bbaw.pdr.solr.SolrCommandException;
import org.bbaw.pdr.solr.SolrPlainQuery;

/**
 * Server tools.
 * @author Fabian Körner
 *
 */
public final class Tools
{

	private static final X509TrustManager[] TRUST_MANAGERS = {
		new X509TrustManager()
		{
			public void checkClientTrusted(final X509Certificate[] certificates, final String string) throws CertificateException
			{  }
			public void checkServerTrusted(final X509Certificate[] certificates, final String string) throws CertificateException
			{  }
			public X509Certificate[] getAcceptedIssuers()
			{
				return null;
			}
		},
	};

	private static final OMFactory OM_FACTORY = OMAbstractFactory.getOMFactory();
	private static final OMNamespace OM_NAMESPACE = OM_FACTORY.createOMNamespace(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.QNAME_PREFIX_ALLIES);

	private Tools()
	{
	}

	/**
	 * Returns the URL encoded <code>text</code>.
	 * @param text Text to be encoded.
	 * @return URL encoded text.
	 */
	public static String cloak(final String text)
	{
		String result = null;
		try
		{
			result = URLEncoder.encode(text, AlliesConstants.HTTP_POST_DATA_ENCODING);
		}
		catch (UnsupportedEncodingException uee)
		{
			/* constant should be OK */
		}
		return result;
	}

	/**
	 * Returns the URL decoded <code>text</code>.
	 * @param text Text to be decoded.
	 * @return URL decoded text.
	 */
	public static String uncloak(final String text)
	{
		String result = null;
		try
		{
			result = URLDecoder.decode(text, AlliesConstants.HTTP_POST_DATA_ENCODING);
		}
		catch (UnsupportedEncodingException uee)
		{
			/* constant should be OK */
		}
		return result;
	}

	/**
	 * Creates the root element of a Allies response.
	 * @param elementName Name of the root element.
	 * @return Root element.
	 */
	public static OMElement createResponseRootElement(final String elementName)
	{
		OMElement result = Tools.OM_FACTORY.createOMElement(elementName, Tools.OM_NAMESPACE);
		result.addAttribute("version", Configuration.getInstance().getVersionNumber(), Tools.OM_NAMESPACE);
		return result;
	}

	/**
	 * Creates an element for use in Allies responses.
	 * @param elementName Name of the element.
	 * @return Response Element.
	 */
	public static OMElement createResponseElement(final String elementName)
	{
		OMElement result = Tools.OM_FACTORY.createOMElement(elementName, Tools.OM_NAMESPACE);
		return result;
	}

	/**
	 * Returns root element of PDR-Allies response element form provided HTTP response.
	 * @param httpResponse HTTP response from server side.
	 * @return Root element of PDR-Allies response.
	 * @throws PDRAlliesClientException Could not get content of PDR-Allies response.
	 */
	public static OMElement getAlliesResponseElement(final HttpResponse httpResponse) throws PDRAlliesClientException
	{
		OMElement soapEnvelopeElement = null;
		String soapResponse = null;
		try
		{
			InputStream inputStream = httpResponse.getEntity().getContent();

			soapResponse = Tools.readInputStream(inputStream);
			soapEnvelopeElement = AXIOMUtil.stringToOM(soapResponse);
		}
		catch (Exception e)
		{
			throw new PDRAlliesClientException("couldn't read allies response", e);
		}

		if ("Envelope".equals(soapEnvelopeElement.getLocalName()))
		{
			OMElement soapBodyElement = soapEnvelopeElement.getFirstElement();
			OMElement responseElement = soapBodyElement.getFirstElement();
			if ("Fault".equals(responseElement.getLocalName()))
			{
				throw new PDRAlliesClientException("found soap fault: " + responseElement.getFirstChildWithName(new QName("faultstring")).getText());
			}
			else
			{
				return responseElement;
			}
		}
		else if ("Reason".equals(soapEnvelopeElement.getLocalName()))
		{
			OMElement textElement = soapEnvelopeElement.getFirstElement();
			throw new PDRAlliesClientException("couldn't get allies response element: " + textElement.getText());
		}
		else
		{
			throw new PDRAlliesClientException("unhandled root element in soap response: " + soapResponse);
		}
	}

	/**
	 * Returns a closeable HTTP client which is HTTPS enabled.
	 * @return Closeable HTTP client.
	 * @throws Exception Could not create HTTP client.
	 */
	public static CloseableHttpClient buildHttpClient() throws Exception
	{
		SSLContext context = SSLContext.getInstance("TLS");
		context.init(null, Tools.TRUST_MANAGERS, null);

		HttpClientBuilder clientBuilder = HttpClientBuilder.create();
		clientBuilder.setSslcontext(context);
		return clientBuilder.build();
	}

	/**
	 * Closes the provided HTTP client.
	 * @param client HTTP client to be closed.
	 */
	public static void closeHttpClient(final CloseableHttpClient client)
	{
		if (null != client)
		{
			try
			{
				client.close();
			}
			catch (Throwable cause)
			{
				throw new IllegalArgumentException("could not close closeable http client", cause);
			}
		}
	}

	/**
	 * Looks for authentication information in the provided element, and checks if
	 * the user is known.
	 * @param inElement Element whoch contains authentication information.
	 * @throws AxisFault Unknown user, or authentication error.
	 */
	public static void checkAuthorization(final OMElement inElement) throws AxisFault
	{
		try
		{
			OMElement credentialsElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_CREDENTIALS));
			Credentials credentials = new Credentials(credentialsElement.getText());
			if (Tools.isKnownUser(credentials))
			{
				return;
			}
			else
			{
				throw new AxisFault("unknown user");
			}
		}
		catch (Exception e)
		{
			throw new AxisFault("authentication error", e);
		}
	}

	private static boolean isKnownUser(final Credentials credentials) throws SolrCommandException
	{
		//FIXME resolve solr dependency
		String userID = credentials.getUserID();
		String md5Password = credentials.getMD5Password();
		String query = "id:" + userID + " AND user.password:" + md5Password;

		SolrQuery solrQuery = new SolrQuery(query);
		QueryResponse response = new SolrPlainQuery(solrQuery).execute();

		Long numFound = response.getResults().getNumFound();

		return 0 < numFound.intValue();
	}

	/**
	 * Reads the provided input stream and returns its content as <code>java.lang.String</code>.
	 * @param inputStream Input stream to read.
	 * @return Content as String.
	 * @throws IOException Could not read input stream.
	 */
	public static String readInputStream(final InputStream inputStream) throws IOException
	{
		StringBuffer sb = new StringBuffer();
		BufferedInputStream bis = new BufferedInputStream(inputStream);
		byte[] buf = new byte[1024];
		int read;
		while (-1 != (read = bis.read(buf)))
		{
			sb.append(new String(buf, 0, read));
		}
		return sb.toString();
	}

}
