/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.RandomAccessFile;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.bbaw.pdr.padre.PadreSingleReadXml;
import org.bbaw.pdr.solr.SolrOptimize;
import org.bbaw.pdr.utils.IDRange;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRObject;
import org.bbaw.pdr.utils.PDRType;
import org.bbaw.pdr.utils.Repository;

/**
 * Commonly useful methods with access to the PDR repository.
 * @author Fabian Körner
 */
public class Utilities
{

	private static final Configuration CONFIGURATION = Configuration.getInstance();

	private static final Lock LOCK_FS = new ReentrantLock();
	private static final long LOCK_FS_TIMEOUT = 60L; /* seconds */

	private static final long LOG_READ_BUFFER = 16384L;

	private static final SimpleDateFormat SOLR_LOG_TIMESTAMP = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
	private static final Pattern PATTERN_CATALINA_OUT_OPTIMIZED_DATE = Pattern.compile("^[0-9]{2}[.][0-9]{2}[.][0-9]{4}[ ][0-9]{2}[:][0-9]{2}[:][0-9]{2}");
	private static final Pattern PATTERN_CATALINA_OUT_OPTIMIZED_INFO = Pattern.compile("[{]optimize[=][}][ ][0-9]+[ ][0-9]+");

	private static final Pattern NCNAME_PATTERN = Pattern.compile("^[a-zA-Z_][0-9a-zA-Z_.-]+$");

	private static final Pattern CATEGORIES_FILENAME_PATTERN = Pattern.compile("^categories[.][a-zA-Z_][0-9a-zA-Z_.-]+?[.]xml$");

	/**
	 * Returns categories of given provider.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with requests categories.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getCategories(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getCategories: ", inElement.toString());

		/* Provider */
		OMElement providerElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROVIDER));
		String provider = Tools.uncloak(providerElement.getText());

		/* Datei lesen */
		String fileContent = "";
		if (Utilities.isValidCategoriesProviderName(provider))
		{
			String filename = Utilities.buildCategoriesFilename(provider);
			CONFIGURATION.logDebug("filename #getCategories: " + filename);

			boolean hasLock = false;
			try
			{
				hasLock = LOCK_FS.tryLock(LOCK_FS_TIMEOUT, TimeUnit.SECONDS);
				fileContent = Utilities.readCategoriesFile(filename);
			}
			catch (Exception e)
			{
				CONFIGURATION.logError("Utilities#getCategories: ", e);
				throw new AxisFault(e.getMessage() + " (Utilities#getCategories)");
			}
			finally
			{
				if (hasLock)
				{
					LOCK_FS.unlock();
				}
			}
		}
		else
		{
			String message = "invalid provider name [" + provider + "]";
			CONFIGURATION.logError("Utilities#getCategories: " + message);
			throw new AxisFault(message + "(Utilities#Categories)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_CATEGORIES_RESPONSE);
		OMElement categoriesElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_CATEGORIES);
		categoriesElement.setText(Tools.cloak(fileContent));
		outElement.addChild(categoriesElement);
		CONFIGURATION.logDebug("output #getCategories: ", outElement.toString());

		return outElement;
	}

	/**
	 * Returns available category providers.
	 * @param inElement Empty Input element.
	 * @return Output element with provider names.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getCategoryProviders(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getCategoryProviders: ", inElement.toString());

		/* Ergebnis */
		File dir = new File(Configuration.getInstance().getCatalinaHome() + "/webapps/ROOT");
		FilenameFilter filter = new FilenameFilter()
		{
			@Override
			public boolean accept(final File dir, final String name)
			{
				Matcher m = CATEGORIES_FILENAME_PATTERN.matcher(name);
				return m.find();
			}
		};
		String[] filenames = dir.list(filter);

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_CATEGORY_PROVIDERS_RESPONSE);
		OMElement providersElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_PROVIDERS);
		OMElement providerElement;
		for (String filename : filenames)
		{
			providerElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_PROVIDER);
			providerElement.setText(Tools.cloak(filename.substring(filename.indexOf('.') + 1, filename.lastIndexOf('.'))));
			providersElement.addChild(providerElement);
		}
		outElement.addChild(providersElement);
		CONFIGURATION.logDebug("output #getCategoryProviders: ", outElement.toString());

		return outElement;
	}

	/**
	 * Returns single PDR object.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with requested object.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getObject(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getObject: ", inElement.toString());

		Tools.checkAuthorization(inElement);

		OMElement idElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_ID));

		/* Ergebnis */
		PDRObject pdrObject;
		try
		{
	    	Identifier id = new Identifier(idElement.getText());
			pdrObject = new PadreSingleReadXml(id.toString()).execute();
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Utilities#getObject: ", e);
			throw new AxisFault(e.getMessage() + " (Utilities#getObject)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_OBJECT_RESPONSE);
		OMElement objectElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECT);
		objectElement.setText(Tools.cloak(pdrObject.getXMLContent()));
		outElement.addChild(objectElement);
		CONFIGURATION.logDebug("output #getObject: ", outElement.toString());

		return outElement;
	}

	/**
	 * Returns all available PDR objects within given range.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with requested objects.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getObjects(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getObjects: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

		OMElement typeElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_TYPE));
		OMElement instanceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_INSTANCE));
		OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));
		OMElement minElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_MIN));
		OMElement maxElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_MAX));

		PDRType type = Identifier.typeFromString(typeElement.getText());
		int instance = Integer.parseInt(instanceElement.getText());
		int project = Integer.parseInt(projectElement.getText());
		int min = Integer.parseInt(minElement.getText());
		int max = Integer.parseInt(maxElement.getText());

		/* Ergebnis */
		List<String> objects = new Vector<String>();
		try
		{
	    	objects = Repository.getObjects(type, instance, project, min, max);
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Utilities#getObjects: ", e);
			throw new AxisFault(e.getMessage() + " (Utilities#getObjects)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_OBJECTS_RESPONSE);
		OMElement objectsElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECTS);
		OMElement objectElement;
		for (String object : objects)
		{
			objectElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_OBJECT);
			objectElement.setText(Tools.cloak(object));
			objectsElement.addChild(objectElement);
		}
		outElement.addChild(objectsElement);
		CONFIGURATION.logDebug("output #getObjects: ", outElement.toString());

		return outElement;
	}

	/**
	 * Returns occupied ID ranges in given boundaries.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with occupied id ranges.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement getOccupiedIDRanges(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #getOccupiedIDRanges: ", inElement.toString());

		//TODO QNames konstant, gilt auch fuer andere Klassen und Methoden
		OMElement typeElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_TYPE));
		OMElement instanceElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_INSTANCE));
		OMElement projectElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROJECT));
		OMElement minElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_MIN));
		OMElement maxElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_MAX));

		PDRType type = Identifier.typeFromString(typeElement.getText());
		int instance = Integer.parseInt(instanceElement.getText());
		int project = Integer.parseInt(projectElement.getText());
		int min = Integer.parseInt(minElement.getText());
		int max = Integer.parseInt(maxElement.getText());

		/* Ergebnis */
		List<IDRange> ranges = new Vector<IDRange>();
		try
		{
	    	ranges = Repository.getOccupiedObjectIDRanges(type, instance, project, min, max);
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Utilities#getOccupiedIDRanges: ", e);
			throw new AxisFault(e.getMessage() + " (Utilities#getOccupiedIDRanges)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_OCCUPIED_ID_RANGES_RESPONSE);
		OMElement rangesElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_RANGES);

		OMElement rangeElement;
		for (IDRange range : ranges)
		{
			rangeElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_RANGE);
			minElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_MIN);
			maxElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_MAX);
			minElement.setText(Integer.toString(range.getLowerBound()));
			maxElement.setText(Integer.toString(range.getUpperBound()));
			rangeElement.addChild(minElement);
			rangeElement.addChild(maxElement);
			rangesElement.addChild(rangeElement);
		}
		outElement.addChild(rangesElement);
		CONFIGURATION.logDebug("output #getOccupiedIDRanges: ", outElement.toString());

		return outElement;
	}

	/**
	 * Returns server-side's version information.
	 * @param inElement Empty input element.
	 * @return Output element with version information.
	 */
	public final OMElement getVersion(final OMElement inElement)
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
    	CONFIGURATION.logDebug("input #getVersion: ", inElement.toString());

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_GET_VERSION_RESPONSE);
    	OMElement value = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_TEXT);
    	value.setText(Configuration.getInstance().getShortInfo());
    	outElement.addChild(value);
    	CONFIGURATION.logDebug("output #getVersion: ", outElement.toString());

		return outElement;
	}

	/**
	 * Looks for information about finished optimisation in Solr's log messages,
	 * newer than the provided timestamp.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Output element with information on successful index optimisation..
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement isIndexOptimized(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #isIndexOptimized: ", inElement.toString());

		OMElement timestampElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_TIMESTAMP));

		Date startDate = null;
		try
		{
			startDate = SOLR_LOG_TIMESTAMP.parse(Tools.uncloak(timestampElement.getText()));
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Utilities#isIndexOptimized: ", e);
			throw new AxisFault(e.getMessage() + " (Utilities#isIndexOptimized)");
		}

		boolean result = false;

		if (startDate != null)
		{
			try
			{
				String filename = Utilities.buildCatalinaOutFilename();
				CONFIGURATION.logDebug("filename #isIndexOptimized: ", filename);
				RandomAccessFile file = new RandomAccessFile(new File(filename), "r");

				if (file.length() <= LOG_READ_BUFFER)
				{
					file.seek(0);
				}
				else
				{
					file.seek(file.length() - LOG_READ_BUFFER);
				}

				Matcher dateMatcher;
				String line;
				while ((line = file.readLine()) != null)
				{
					dateMatcher = PATTERN_CATALINA_OUT_OPTIMIZED_DATE.matcher(line);
					if (dateMatcher.find())
					{
						Date lineDate = SOLR_LOG_TIMESTAMP.parse(dateMatcher.group());
						if (lineDate.compareTo(startDate) >= 0)
						{
							Matcher infoMatcher;
							while ((line = file.readLine()) != null)
							{
								infoMatcher = PATTERN_CATALINA_OUT_OPTIMIZED_INFO.matcher(line);
	    						if (infoMatcher.find())
	    						{
									result = true;
									break;
	    						}
							}
							if (result)
							{
								break;
							}
						}
					}
				}
				file.close();
	    	}
	    	catch (Exception e)
	    	{
				CONFIGURATION.logError("Utilities#isIndexOptimized: ", e);
				throw new AxisFault(e.getMessage() + " (Utilities#isIndexOptimized)");
	    	}
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_IS_INDEX_OPTIMIZED_RESPONSE);
		OMElement valueElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_VALUE);
		valueElement.setText(Tools.cloak(Boolean.toString(result)));
		outElement.addChild(valueElement);
		CONFIGURATION.logDebug("output #isIndexOptimized: ", outElement.toString());

		return outElement;
	}

	/**
	 * Start Solr's optimisation and returns current timestamp.
	 * @param inElement Empty input element.
	 * @return Output element with timestamp from start of the optimisation.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement optimizeIndex(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #optimizeIndex: ", inElement.toString());

		/* Rückgabe: Serverzeit */
		String timestamp = SOLR_LOG_TIMESTAMP.format(new Date(System.currentTimeMillis()));

		/* Optimierung anstoßen */
		try
		{
			new SolrOptimize().execute();
		}
		catch (Exception e)
		{
			CONFIGURATION.logError("Utilities#optimizeIndex: ", e);
			throw new AxisFault(e.getMessage() + " (Utilities#optimizeIndex)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_OPTIMIZE_INDEX_RESPONSE);
		OMElement timestampElement = Tools.createResponseElement(AlliesConstants.OMCHILD_LNAME_TIMESTAMP);

		timestampElement.setText(Tools.cloak(timestamp));
		outElement.addChild(timestampElement);
		CONFIGURATION.logDebug("output #optimizeIndex: ", outElement.toString());

		return outElement;
	}

	/**
	 * Deletes categories of given provider.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Empty output element.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement removeCategories(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
		inElement.build();
		inElement.detach();
		CONFIGURATION.logDebug("input #removeCategories: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

		OMElement providerElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROVIDER));
		String provider = Tools.uncloak(providerElement.getText());

		if (Utilities.isValidCategoriesProviderName(provider))
		{
			String filename = Utilities.buildCategoriesFilename(provider);
			CONFIGURATION.logDebug("filename #removeCategories: " + filename);
			new File(filename).delete();
		}
		else
		{
			String message = "invalid provider name [" + provider + "]";
			CONFIGURATION.logError("Utilities#removeCategories: " + message);
			throw new AxisFault(message + "(Utilities#removeCategories)");
		}

		/* Ausgabe */
		OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_REMOVE_CATEGORIES_RESPONSE);
		CONFIGURATION.logDebug("output #removeCategories: ", outElement.toString());

		return outElement;
	}

	/**
	 * Stores categories of given provider.
	 * @param inElement Input element with necessary parameters and values.
	 * @return Empty output element.
	 * @throws AxisFault Error occured during processing.
	 */
	public final OMElement setCategories(final OMElement inElement) throws AxisFault
	{
		/* Eingabe */
    	inElement.build();
    	inElement.detach();
    	CONFIGURATION.logDebug("input #setCategories: ", inElement.toString());

		/* Authorisierung */
		Tools.checkAuthorization(inElement);

    	OMElement categoriesElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_CATEGORIES));
    	OMElement providerElement = inElement.getFirstChildWithName(new QName(AlliesConstants.OMCHILD_LNAME_PROVIDER));

    	String categories = Tools.uncloak(categoriesElement.getText());
    	String provider = Tools.uncloak(providerElement.getText());

    	if (Utilities.isValidCategoriesProviderName(provider))
    	{
    		boolean hasLock = false;
    		try
    		{
    			hasLock = LOCK_FS.tryLock(LOCK_FS_TIMEOUT, TimeUnit.SECONDS);

        		String filename = Utilities.buildCategoriesFilename(provider);
        		CONFIGURATION.logDebug("filename #setCategories: " + filename);

        		Utilities.writeCategoriesFile(filename, categories);
    		}
    		catch (Exception e)
    		{
    			CONFIGURATION.logError("Utilities#setCategories: ", e);
    			throw new AxisFault(e.getMessage() + " (Utilities#setCategories)");
    		}
    		finally
    		{
    			if (hasLock)
    			{
    				LOCK_FS.unlock();
    			}
    		}
    	}
		else
		{
			String message = "invalid provider name [" + provider + "]";
			CONFIGURATION.logError("Utilities#setCategories: " + message);
			throw new AxisFault(message + "(Utilities#setCategories)");
		}

    	/* Ausgabe */
    	OMElement outElement = Tools.createResponseRootElement(AlliesConstants.OMCHILD_UTILITIES_SET_CATEGORIES_RESPONSE);
    	CONFIGURATION.logDebug("output #setCategories: ", outElement.toString());

    	return outElement;
	}

	private static String buildCatalinaOutFilename()
	{
		return 	Configuration.getInstance().getCatalinaHome() + "/logs/catalina.out";
	}

	private static String buildCategoriesFilename(final String provider)
	{
		StringBuilder sb = new StringBuilder();
		sb.append(Configuration.getInstance().getCatalinaHome());
		sb.append("/webapps/ROOT/categories.");
		sb.append(provider);
		sb.append(".xml");
		return sb.toString();
	}

	private static boolean isValidCategoriesProviderName(final String provider)
	{
		if (provider == null)
		{
			return false;
		}
		else if (provider.length() == 0)
		{
			return false;
		}
		else
		{
    		Matcher matcher = NCNAME_PATTERN.matcher(provider);
    		return matcher.find();
		}
	}

	private static String readCategoriesFile(final String filename)
	throws IOException
	{
		FileReader fileReader = new FileReader(new File(filename));
		LineNumberReader lnr = new LineNumberReader(fileReader);
		StringBuilder sb = new StringBuilder();
		String line;
		while ((line = lnr.readLine()) != null)
		{
			sb.append(line);
			sb.append('\n');
		}
		lnr.close();
		return sb.substring(0, sb.lastIndexOf("\n"));
	}

	private static void writeCategoriesFile(final String filename, final String categories)
	throws IOException
	{
		FileWriter fw = new FileWriter(new File(filename));
		BufferedWriter bw = new BufferedWriter(fw);
		bw.write(categories);
		bw.flush();
		bw.close();
	}
}
