/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import org.apache.commons.codec.digest.DigestUtils;

/**
 * Holds information about how to access PDR's Apache Axis2 interface.
 * @author Fabian Körner
 */
public final class Configuration
{

	private String _axis2BaseURL;
	private String _pdrUserID;
	private String _pdrUserPasswordMD5;

	private static final Configuration INSTANCE = new Configuration();

	private Configuration()
	{
	}

	/**
	 * Singleton: returns a reference to this type's unique instance.
	 * @return Instance.
	 */
	public static Configuration getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Returns the configured URL of PDR's Apache Axis2 interface as String.
	 * @return Apache Axis2 base URL.
	 */
	public synchronized String getAxis2BaseURL()
	{
		return _axis2BaseURL;
	}

	/**
	 * Returns the PDR user ID of the currently configured user.
	 * @return PDR user ID.
	 */
	public synchronized String getPDRUserID()
	{
		return _pdrUserID;
	}

	/**
	 * Returns md5 hash of user's password.
	 * @return MD5 hash of password.
	 */
	public synchronized String getMD5Password()
	{
		return _pdrUserPasswordMD5;
	}

	protected synchronized String getCredentials()
	{
		return _pdrUserID + "#" + _pdrUserPasswordMD5;
	}

	/**
	 * Sets the URL of PDR's Apache Axis2 interface, which will
	 * be used for all connections from now on.
	 * (Format: <code>https://hostname:port/axis2</code>).
	 * @param axis2BaseURL URL of PDR's Apache Axis2 interface.
	 */
	public synchronized void setAxis2BaseURL(final String axis2BaseURL)
	{
		_axis2BaseURL = axis2BaseURL;
	}

	/**
	 * Sets current User's PDR user ID and password. This information
	 * will be used for authentication with server side.
	 * @param pdrUserID PDR user ID.
	 * @param password User's password.
	 */
	public synchronized void setPDRUser(final String pdrUserID, final String password)
	{
		_pdrUserID = pdrUserID;
		_pdrUserPasswordMD5 = DigestUtils.md5Hex(password);
	}
}
