/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.SSLContext;
import javax.net.ssl.X509TrustManager;
import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.util.AXIOMUtil;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicHeader;
import org.bbaw.pdr.allies.AlliesConstants;
import org.bbaw.pdr.allies.error.PDRAlliesClientException;
import org.bbaw.pdr.allies.error.PDRIndexException;
import org.bbaw.pdr.allies.error.PDRStorageException;
import org.bbaw.pdr.allies.error.PDRValidationException;

/**
 * Client tools.
 * @author Fabian Körner
 */
public final class Tools
{

	private static final Header HEADER_POST_REQUEST_CONTENT_TYPE = new BasicHeader("Content-Type", "text/xml");
	private static final String POST_REQUEST_ENCODING = "UTF-8";

	private static final X509TrustManager[] TRUST_MANAGERS = {
		new X509TrustManager()
		{
			public void checkClientTrusted(final X509Certificate[] certificates, final String string) throws CertificateException
			{  }
			public void checkServerTrusted(final X509Certificate[] certificates, final String string) throws CertificateException
			{  }
			public X509Certificate[] getAcceptedIssuers()
			{
				return null;
			}
		},
	};

	/* Hidden default constructor. */
	private Tools()
	{
	}

	/**
	 * Returns the URL decoded <code>text</code>.
	 * @param text Text to be decoded.
	 * @return URL decoded text.
	 */
	public static String uncloak(final String text)
	{
		String result = null;
		try
		{
			result = URLDecoder.decode(text, AlliesConstants.HTTP_POST_DATA_ENCODING);
		}
		catch (UnsupportedEncodingException uee)
		{
			/* constant should be OK */
		}
		return result;
	}

	/**
	 * Returns the URL encoded <code>text</code>.
	 * @param text Text to be encoded.
	 * @return URL encoded text.
	 */
	public static String cloak(final String text)
	{
		String result = null;
		try
		{
			result = URLEncoder.encode(text, AlliesConstants.HTTP_POST_DATA_ENCODING);
		}
		catch (UnsupportedEncodingException uee)
		{
			/* constant should be OK */
		}
		return result;
	}

	/**
	 * Returns a closeable HTTP(S) client.
	 * @return Closeable HTTP(S) client.
	 * @throws PDRAlliesClientException Could not create HTTP client.
	 */
	public static CloseableHttpClient buildHttpClient() throws PDRAlliesClientException
	{
		try
		{
			SSLContext context = SSLContext.getInstance("TLS");
			context.init(null, Tools.TRUST_MANAGERS, null);

			HttpClientBuilder clientBuilder = HttpClientBuilder.create();
			clientBuilder.setSslcontext(context);
			return clientBuilder.build();
		}
		catch (Exception cause)
		{
			throw new PDRAlliesClientException(cause);
		}
	}

	/**
	 * Closes the provided HTTP client.
	 * @param client Closeable HTTP client.
	 */
	public static void closeHttpClient(final CloseableHttpClient client)
	{
		if (null != client)
		{
			try
			{
				client.close();
			}
			catch (Throwable cause)
			{
				throw new IllegalArgumentException("could not close closeable http client", cause);
			}
		}
	}

	/**
	 * Build a HTTP POST request which can be sent to server side.
	 * @param methodURL Complete URL of the used method.
	 * @param dataPart Data part of the request.
	 * @return HTTP POST request which can be sent to server side.
	 */
	public static HttpUriRequest createHttpPostRequest(final String methodURL, final String dataPart)
	{
		HttpPost post = new HttpPost(methodURL);
		post.addHeader(Tools.HEADER_POST_REQUEST_CONTENT_TYPE);
		try
		{
			post.setEntity(new StringEntity(dataPart, Tools.POST_REQUEST_ENCODING));
		}
		catch (UnsupportedEncodingException cause)
		{
			throw new IllegalArgumentException(cause);
		}
		return post;
	}

	/**
	 * Retruns the PDR-Allies response Element from incoming HTTP response.
	 * @param httpResponse Incoming HTTP response.
	 * @return PDR-Allies response Element.
	 * @throws PDRAlliesClientException Could not get response element.
	 */
	public static OMElement getAlliesResponseElement(final HttpResponse httpResponse) throws PDRAlliesClientException
	{
		//TODO refactor method
		OMElement soapEnvelopeElement = null;
		String soapResponse = null;
		try
		{
			InputStream inputStream = httpResponse.getEntity().getContent();

			soapResponse = Tools.readInputStream(inputStream);
			soapEnvelopeElement = AXIOMUtil.stringToOM(soapResponse);
		}
		catch (Exception e)
		{
			throw new PDRAlliesClientException("couldn't read allies response", e);
		}

		if ("Envelope".equals(soapEnvelopeElement.getLocalName()))
		{
			OMElement soapBodyElement = soapEnvelopeElement.getFirstElement();
			OMElement responseElement = soapBodyElement.getFirstElement();
			if ("Fault".equals(responseElement.getLocalName()))
			{
				String faultString = responseElement.getFirstChildWithName(new QName("faultstring")).getText();
				if (faultString.startsWith("[STORAGE]-"))
				{
					throw new PDRStorageException(faultString.substring(10));
				}
				if (faultString.startsWith("[INDEX]-"))
				{
					throw new PDRIndexException(faultString.substring(8));
				}
				if (faultString.startsWith("[VALIDATION]-"))
				{
					throw new PDRValidationException(faultString.substring(13));
				}
				else
				{
					throw new PDRAlliesClientException("found soap fault: " + faultString);
				}
			}
			else
			{
				return responseElement;
			}
		}
		else if ("Reason".equals(soapEnvelopeElement.getLocalName()))
		{
			OMElement textElement = soapEnvelopeElement.getFirstElement();
			throw new PDRAlliesClientException("couldn't get allies response element: " + textElement.getText());
		}
		else
		{
			throw new PDRAlliesClientException("unhandled root element in soap response: " + soapResponse);
		}
	}

	/**
	 * Reads the provided input stream and returns its content as <code>java.lang.String</code>.
	 * @param inputStream Input stream to read.
	 * @return Content as String.
	 * @throws IOException Could not read input stream.
	 */
	public static String readInputStream(final InputStream inputStream) throws IOException
	{
		StringBuffer sb = new StringBuffer();
		BufferedInputStream bis = new BufferedInputStream(inputStream);
		byte[] buf = new byte[1024];
		int read;
		while (-1 != (read = bis.read(buf)))
		{
			sb.append(new String(buf, 0, read));
		}
		return sb.toString();
	}

	/**
	 * Sets Axis2 base URL.
	 * @param url Axis2 base URL.
	 */
	public static void setAxis2Base(final URL url)
	{
		String axis2Base = url.toString();
		Configuration.getInstance().setAxis2BaseURL(axis2Base);
	}
}
