/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.impl.client.CloseableHttpClient;
import org.bbaw.pdr.allies.AlliesConstants;
import org.bbaw.pdr.allies.error.PDRAlliesClientException;

/**
 * Common helpers.
 * @author Fabian Körner
 */
public final class Utilities
{

	private static final Pattern PATTERN_VERSION_NUMBER = Pattern.compile("(\\d+)[.](\\d+)[.](\\d+)");

	private Utilities()
	{
	}

	/**
	 * Returns server side's version information.
	 * @return Server-side version number.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static String getServerVersion() throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/getVersion";
			String dataPart = DataParts.createGetVersion();

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);

			OMElement textElement = null;
			Iterator<?> i = alliesResponseElement.getChildElements();
			OMElement currentElement;
			while (i.hasNext())
			{
				currentElement = (OMElement) i.next();
				if ("Text".equals(currentElement.getLocalName()))
				{
					textElement = currentElement;
					break;
				}
			}
			return Tools.uncloak(textElement.getText());
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Checks if server side version equals or is greater than a given version
	 * number (pattern: 0.0.0).
	 * @param expected Expected version number.
	 * @return Version number equals or is greater than the expected number.
	 * @throws PDRAlliesClientException Fehlerfall.
	 */
	public static boolean checkServerVersion(final String expected) throws PDRAlliesClientException
	{
		String actual = Utilities.getServerVersion();

		int actualValue = Utilities.versionToInt(actual);
		int expectedValue = Utilities.versionToInt(expected);

		return actualValue >= expectedValue;
	}

	/**
	 * Returns the categories of a given provider.
	 * @param provider Name of the provider.
	 * @return Categories.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static String getCategories(final String provider) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			String result = null;

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/getCategories";
			String dataPart = DataParts.createGetCategories(provider);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);

			OMElement categoriesElement = alliesResponseElement.getFirstElement();

			result = Tools.uncloak(categoriesElement.getText());

			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Returns available category providers.
	 * @return Category providers.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static List<String> getCategoryProviders() throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			List<String> result = new Vector<String>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/getCategoryProviders";
			String dataPart = DataParts.createGetCategoryProviders();

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement providersElement = alliesResponseElement.getFirstElement();
			Iterator<?> providersIterator = providersElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_PROVIDER,
							AlliesConstants.QNAME_PREFIX_ALLIES));

			OMElement providerElement;
			while (providersIterator.hasNext())
			{
				providerElement = (OMElement) providersIterator.next();
				result.add(Tools.uncloak(providerElement.getText()));
			}
			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Returns occupied ID ranges in given boundaries.
	 * @param type Object type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound of specified range.
	 * @param upperBound Upper bound of specified range.
	 * @return Fitting ID ranges.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static List<IDRange> getOccupiedObjectIDRanges(final PDRType type, final int instance, final int project, final int lowerBound,
			final int upperBound) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			List<IDRange> result = new Vector<IDRange>();
			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/getOccupiedIDRanges";
			String dataPart = DataParts.createGetOccupiedObjectIDRanges(type, instance, project, lowerBound, upperBound);

			httpClient = Tools.buildHttpClient();
			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement rangesElement = alliesResponseElement.getFirstElement();
			Iterator<?> rangesIterator = rangesElement.getChildrenWithName(new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_RANGE,
					AlliesConstants.QNAME_PREFIX_ALLIES));
			OMElement rangeElement;
			OMElement minElement;
			OMElement maxElement;
			while (rangesIterator.hasNext())
			{
				rangeElement = (OMElement) rangesIterator.next();
				minElement = rangeElement.getFirstChildWithName(new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
						AlliesConstants.OMCHILD_LNAME_MIN, AlliesConstants.QNAME_PREFIX_ALLIES));
				maxElement = rangeElement.getFirstChildWithName(new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
						AlliesConstants.OMCHILD_LNAME_MAX, AlliesConstants.QNAME_PREFIX_ALLIES));
				result.add(new IDRange(type, instance, project, Integer.parseInt(minElement.getText()), Integer.parseInt(maxElement.getText())));
			}
			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Deletes categories of given provider on server-side.
	 * @param provider Provider name.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static void removeCategories(final String provider) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/removeCategories";
			String dataPart = DataParts.createRemoveCategories(provider);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			httpClient.execute(post);
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Stores categories of given provider on server-side.
	 * @param categories Categories.
	 * @param provider Provider name.
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static void setCategories(final String categories, final String provider) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/setCategories";
			String dataPart = DataParts.createSetCategories(provider, categories);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			httpClient.execute(post);
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	private static int versionToInt(final String version) throws PDRAlliesClientException
	{
		Matcher m = PATTERN_VERSION_NUMBER.matcher(version);
		if (m.find())
		{
			StringBuilder sb = new StringBuilder();
			sb.append(m.group(1));
			sb.append(m.group(2));
			sb.append(m.group(3));
			return Integer.parseInt(sb.toString());
		}
		else
		{
			throw new PDRAlliesClientException("malformed version number: " + version);
		}
	}

	/**
	 * Returns all available PDR objects within given range.
	 * @param type Object type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound of specified range.
	 * @param upperBound Upper bound of specified range.
	 * @return Fitting PDR objects (XML).
	 * @throws PDRAlliesClientException An error occured during processing the request.
	 */
	public static Vector<String> getObjects(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound)
			throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			final Vector<String> result = new Vector<String>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Utilities/getObjects";
			String dataPart = DataParts.createGetObjects(type, instance, project, lowerBound, upperBound);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement objectsElement = alliesResponseElement.getFirstElement();
			Iterator<?> objectsIterator = objectsElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_OBJECT,
							AlliesConstants.QNAME_PREFIX_ALLIES));

			OMElement objectElement;
			while (objectsIterator.hasNext())
			{
				objectElement = (OMElement) objectsIterator.next();
				result.add(Tools.uncloak(objectElement.getText()));
			}
			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}
}
