/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client.tests;

import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.LineNumberReader;
import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.solr.client.solrj.SolrServer;
import org.apache.solr.client.solrj.impl.HttpSolrServer;
import org.bbaw.pdr.allies.client.Configuration;
import org.bbaw.pdr.allies.client.Identifier;
import org.bbaw.pdr.allies.client.Repository;
import org.bbaw.pdr.padre.PadreConfiguration;
import org.bbaw.pdr.padre.client.PadreClient;
import org.bbaw.pdr.solr.SolrConfiguration;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * @author Fabian Körner
 */
public class TestAEBackupBertholdLokal
{

	private static final org.bbaw.pdr.allies.client.tests.Configuration TEST_CONFIG = new ConfigurationFromFile();

	private static PadreClient _padreClient;
	private static SolrServer _solrServer;

	private static List<String> _validReferences;
	private static List<String> _validPersons;
	private static List<String> _validAspects;

	private static List<String> _invalidReferences;
	private static List<String> _invalidPersons;
	private static List<String> _invalidAspects;

	/**
	 * @throws Exception Exception.
	 */
	@BeforeClass
	public static void setUp() throws Exception
	{
		Configuration.getInstance().setAxis2BaseURL(TEST_CONFIG.getAxis2BaseUri());
		Configuration.getInstance().setPDRUser("pdrUo.001.002.000000001", TEST_CONFIG.getPDRUserPW(1, 2, 1));

		PadreConfiguration.getInstance().setCredentials(URI.create(TEST_CONFIG.getPadreUri()), TEST_CONFIG.getPadreUsername(), TEST_CONFIG.getPadrePassword());
		SolrConfiguration.getInstance().setBaseURI(URI.create(TEST_CONFIG.getSolrUri()));
		_solrServer = SolrConfiguration.getInstance().getSolrServer();
		_padreClient = PadreConfiguration.getInstance().getPadreClient();
		_padreClient.createStore("321");
		_padreClient.createStore("322");

		TestAEBackupBertholdLokal.prepareTestData();
	}

	/**
	 * @throws Exception Exception.
	 */
	@AfterClass
	public static void tearDown() throws Exception
	{
		_padreClient.deleteStore("321");
		_padreClient.deleteStore("322");
		_solrServer.deleteByQuery("id.project:801");
		_solrServer.deleteByQuery("id.project:802");
		_solrServer.commit();
		((HttpSolrServer) _solrServer).shutdown();
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIngestValidObjects() throws Exception
	{
		Map<Identifier, Identifier> mapReferences = Repository.ingestObjects(1, 802, _validReferences);
		assertSame("output map must be of same size as input map", _validReferences.size(), mapReferences.size());
		this.assertPairStrings(mapReferences,
				"pdrRo.001.802.100000001=pdrRo.001.802.000000001",
				"pdrRo.001.802.100000003=pdrRo.001.802.000000002",
				"pdrRo.001.802.100000004=pdrRo.001.802.000000003"
				);

		_validPersons = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_validPersons, mapReferences);
		_validAspects = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_validAspects, mapReferences);

		Map<Identifier, Identifier> mapPersons = Repository.ingestObjects(1, 802, _validPersons);
		assertSame("output map must be of same size as input map", _validPersons.size(), mapPersons.size());
		this.assertPairStrings(mapPersons,
				"pdrPo.001.802.100000001=pdrPo.001.802.000000001",
				"pdrPo.001.802.100000004=pdrPo.001.802.000000003",
				"pdrPo.001.802.100000013=pdrPo.001.802.000000011",
				"pdrPo.001.802.100000021=pdrPo.001.802.000000019",
				"pdrPo.001.802.100000030=pdrPo.001.802.000000028"
				);

		_validAspects = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_validAspects, mapPersons);

		Map<Identifier, Identifier> mapAspects = Repository.ingestObjects(1, 802, _validAspects);
		assertSame("output map must be of same size as input map", _validAspects.size(), mapAspects.size());
		this.assertPairStrings(mapAspects,
				"pdrAo.001.802.100000001=pdrAo.001.802.000000001",
				"pdrAo.001.802.100000035=pdrAo.001.802.000000020",
				"pdrAo.001.802.100000062=pdrAo.001.802.000000047",
				"pdrAo.001.802.100000089=pdrAo.001.802.000000073",
				"pdrAo.001.802.100000106=pdrAo.001.802.000000089"
				);

	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = org.bbaw.pdr.allies.error.PDRAlliesClientException.class)
	public final void testIngestInValidObjects() throws Exception
	{
		Map<Identifier, Identifier> mapReferences = Repository.ingestObjects(1, 801, _invalidReferences);
//		System.out.println(_invalidReferences);
		assertSame("output map must be of same size as input map", _invalidReferences.size(), mapReferences.size());
		this.assertPairStrings(mapReferences,
				"pdrRo.001.801.100000001=pdrRo.001.801.000000001",
				"pdrRo.001.801.100000003=pdrRo.001.801.000000002",
				"pdrRo.001.801.100000004=pdrRo.001.801.000000003"
				);

		_invalidPersons = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_invalidPersons, mapReferences);
		_invalidAspects = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_invalidAspects, mapReferences);

		Map<Identifier, Identifier> mapPersons = Repository.ingestObjects(1, 801, _invalidPersons);
		assertSame("output map must be of same size as input map", _invalidPersons.size(), mapPersons.size());
		this.assertPairStrings(mapPersons,
				"pdrPo.001.801.100000001=pdrPo.001.801.000000001",
				"pdrPo.001.801.100000004=pdrPo.001.801.000000003",
				"pdrPo.001.801.100000013=pdrPo.001.801.000000011",
				"pdrPo.001.801.100000021=pdrPo.001.801.000000019",
				"pdrPo.001.801.100000030=pdrPo.001.801.000000028"
				);

		_invalidAspects = TestAEBackupBertholdLokal.replaceTemporalIdentifiers(_invalidAspects, mapPersons);

//		System.out.println(_invalidAspects);
		Map<Identifier, Identifier> mapAspects = Repository.ingestObjects(1, 801, _invalidAspects);
		assertSame("output map must be of same size as input map", _invalidAspects.size(), mapAspects.size());
		this.assertPairStrings(mapAspects,
				"pdrAo.001.801.100000001=pdrAo.001.801.000000001",
				"pdrAo.001.801.100000035=pdrAo.001.801.000000020",
				"pdrAo.001.801.100000062=pdrAo.001.801.000000047",
				"pdrAo.001.801.100000089=pdrAo.001.801.000000073",
				"pdrAo.001.801.100000106=pdrAo.001.801.000000089"
				);

	}

	private void assertPairStrings(final Map<Identifier, Identifier> map, final String... expectedPairStrings) throws Exception
	{
		String mapString = map.toString();

		for (String pairString : expectedPairStrings)
		{
			assertTrue("missing expected pair string [" + pairString + "] in " + mapString, mapString.contains(pairString));
		}
	}

	private static List<String> replaceTemporalIdentifiers(final List<String> objects, final Map<Identifier, Identifier> map)
	{
		List<String> result = new Vector<String>();
		for (String object : objects)
		{
			for (Entry<Identifier, Identifier> entry : map.entrySet())
			{
				object = object.replaceAll(entry.getKey().toString(), entry.getValue().toString());
			}
			result.add(object);
		}
		return result;
	}

	private static void prepareTestData() throws Exception
	{
		// load Objects
		_invalidReferences = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/reference/", false);
		_invalidPersons = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/person/", false);
		_invalidAspects = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/aspect/", false);
		// load fixed Objects
		_validReferences = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/reference/", true);
		_validPersons = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/person/", true);
		_validAspects = TestAEBackupBertholdLokal.readObjectsFromDirectory("./test-data/AEBackup-Berthold-Lokal/aspect/", true);
		// change project id
		_invalidReferences = TestAEBackupBertholdLokal.changeInstanceAndProject(_invalidReferences, 3, 1, 1, 801);
		_invalidPersons = TestAEBackupBertholdLokal.changeInstanceAndProject(_invalidPersons, 3, 1, 1, 801);
		_invalidAspects = TestAEBackupBertholdLokal.changeInstanceAndProject(_invalidAspects, 3, 1, 1, 801);
		_validReferences = TestAEBackupBertholdLokal.changeInstanceAndProject(_validReferences, 3, 1, 1, 802);
		_validPersons = TestAEBackupBertholdLokal.changeInstanceAndProject(_validPersons, 3, 1, 1, 802);
		_validAspects = TestAEBackupBertholdLokal.changeInstanceAndProject(_validAspects, 3, 1, 1, 802);
		// sort to get reproducible persistent IDs
		Collections.sort(_invalidReferences);
		Collections.sort(_invalidPersons);
		Collections.sort(_invalidAspects);
		Collections.sort(_validReferences);
		Collections.sort(_validPersons);
		Collections.sort(_validAspects);
	}

	private static List<String> readObjectsFromDirectory(final String dirname, final boolean fixObjects) throws Exception
	{
		File dir = new File(dirname);
		String[] filenames = dir.list(new FilenameFilter() {

			@Override
			public boolean accept(final File arg0, final String arg1)
			{
				String filename = arg1;
				return filename.endsWith("xml");
			}
		});

		List<String> result = new Vector<String>();

		LineNumberReader lnr;
		StringBuilder sb;
		String line;
		for (String filename : filenames)
		{
			lnr = new LineNumberReader(new FileReader(dirname + filename));
			sb = new StringBuilder();
			while (null != (line = lnr.readLine()))
			{
				if (fixObjects)
				{
					sb.append(TestAEBackupBertholdLokal.fixObjects(line));
				}
				else
				{
					sb.append(line);
				}
			}
			result.add(sb.toString());
			lnr.close();
		}

		return result;
	}

	private static String fixObjects(final String line)
	{
		String result = line;
		result = result.replaceAll("<aodl:relation provider=\"PDR\">aspect_of</aodl:relation>",
				"<aodl:relation provider=\"PDR\" object=\"pdrPo.003.001.100000001\">aspect_of</aodl:relation>");
		result = result.replaceAll("<aodl:validationStm>", "<aodl:validationStm authority=\"pdrUo.003.001.000000001\">");
		result = result.replaceAll("<aodl:reference/>", "<aodl:reference>pdrRo.003.001.100000001</aodl:reference>");
		result = result.replaceAll("(mods|podl|aodl):", "");
		result = result.replaceAll("xmlns:(mods|podl|aodl)", "xmlns");
		return result;
	}

	private static List<String> changeInstanceAndProject(final List<String> objects,
			final int currentInstance, final int currentProject, final int newInstance, final int newProject) throws Exception
	{
		List<String> result = new Vector<String>();

		String currentInstancePart = TestAEBackupBertholdLokal.addLeadingZeros(currentInstance, 3);
		String currentProjectPart = TestAEBackupBertholdLokal.addLeadingZeros(currentProject, 3);

		String newInstancePart = TestAEBackupBertholdLokal.addLeadingZeros(newInstance, 3);
		String newProjectPart = TestAEBackupBertholdLokal.addLeadingZeros(newProject, 3);

		Pattern oldIDPattern = Pattern.compile("(pdr[APRU]o[.])" + currentInstancePart + "([.])" + currentProjectPart + "([.]\\d{9})");

		String replacement;
		Matcher matcher;
		for (String object : objects)
		{
			while ((matcher = oldIDPattern.matcher(object)).find())
			{
				replacement = matcher.group(1) + newInstancePart + matcher.group(2) + newProjectPart + matcher.group(3);
				object = matcher.replaceFirst(replacement);
			}
			result.add(object);
		}

		return result;
	}

	private static String addLeadingZeros(final int i, final int length)
	{
		StringBuilder sb = new StringBuilder(Integer.toString(i));

		while (sb.length() < length)
		{
			sb.insert(0, '0');
		}

		return sb.toString();
	}

}
