/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client.tests;

import static org.junit.Assert.assertEquals;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.bbaw.pdr.allies.client.Configuration;
import org.bbaw.pdr.allies.client.Tools;
import org.junit.Test;

/**
 * @author Fabian Körner
 */
public class TestClientTools
{

	private static final 	String[][] TEST_CASES_URL_ENCODING = {
			{"", ""},
			{" ", "+"},
			{"%", "%25"},
			{"</>=\"", "%3C%2F%3E%3D%22"},
			{"äöüßÄÖÜ", "%C3%A4%C3%B6%C3%BC%C3%9F%C3%84%C3%96%C3%9C"},
			{":", "%3A"},       // Reservierte Zeichen laut RFC 3986 (http://tools.ietf.org/html/rfc3986)
			{"/", "%2F"},       // Brauchbarer Konverter: http://www.ulimatbach.de/links/url_decoder.html
			{"?", "%3F"},
			{"#", "%23"},
			{"[]", "%5B%5D"},
			{"@", "%40"},
			{"!", "%21"},
			{"$", "%24"},
			{"&", "%26"},
			{"'", "%27"},
			{"()", "%28%29"},
			{"*", "*"},           // ?
			{"+", "%2B"},
			{",", "%2C"},
			{";", "%3B"},
			{"=", "%3D"},
			{"{}", "%7B%7D"},    // weitere unsichere Zeichen, falls nicht schon geprüft
			{"|", "%7C"},        // (Quelle: http://www.blooberry.com/indexdot/html/topics/urlencoding.htm)
			{"^", "%5E"},
			{"~", "%7E"},
			{"`", "%60"},
	};

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testCloakByURLEncoding() throws Exception
	{
		String decoded;
		String encoded;
		for (String[] testCase : TestClientTools.TEST_CASES_URL_ENCODING)
		{
			decoded = testCase[0];
			encoded = testCase[1];
			assertEquals(encoded, Tools.cloak(decoded));
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testUncloakByURLEncoding() throws Exception
	{
		String encoded;
		String decoded;
		for (String[] testCase : TestClientTools.TEST_CASES_URL_ENCODING)
		{
			encoded = testCase[1];
			decoded = testCase[0];
			assertEquals(decoded, Tools.uncloak(encoded));
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testReadInputStream() throws Exception
	{
		StringBuilder sb = new StringBuilder();
		for (String[] testCase : TestClientTools.TEST_CASES_URL_ENCODING)
		{
			sb.append(testCase[0]);
			sb.append(testCase[1]);
		}
		String expected = sb.toString();

		InputStream inputStream = new ByteArrayInputStream(expected.getBytes("UTF-8"));

		String actual = Tools.readInputStream(inputStream);

		assertEquals(expected, actual);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testAxis2BaseURL() throws Exception
	{
		String expected = "http://localhost:8080/lalala";
		Tools.setAxis2Base(new URL(expected));

		assertEquals(expected, Configuration.getInstance().getAxis2BaseURL());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = MalformedURLException.class)
	public final void testInvalidAxis2BaseURL() throws Exception
	{
		String expected = "http localhost:8080/lalala";
		Tools.setAxis2Base(new URL(expected));

		assertEquals(expected, Configuration.getInstance().getAxis2BaseURL());
	}
}
