/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client.tests;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertFalse;

import org.bbaw.pdr.allies.client.Identifier;
import org.bbaw.pdr.allies.client.PDRType;
import org.bbaw.pdr.allies.error.InvalidIdentifierException;
import org.junit.Test;

/**
 * @author Fabian Körner
 */
public class TestIdentifier
{

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringToString() throws Exception
	{
		String pdrID = "pdrAo.001.002.000000003";
		Identifier identifer = new Identifier(pdrID);
		assertEquals(pdrID, identifer.toString());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringEquals() throws Exception
	{
		String pdrID = "pdrPo.004.005.000000006";
		Identifier identifier = new Identifier(pdrID);
		assertTrue(identifier.equals(new Identifier(pdrID)));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringNotEqual() throws Exception
	{
		String pdrID = "pdrPo.004.005.000000006";
		Identifier identifier = new Identifier(pdrID);
		assertFalse(identifier.equals(new Identifier("pdrPo.004.005.000000060")));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringGetType() throws Exception
	{
		String pdrID = "pdrRo.007.008.000000009";
		Identifier identifier = new Identifier(pdrID);
		assertEquals(PDRType.REFERENCE, identifier.getType());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringGetNumbers() throws Exception
	{
		String pdrID = "pdrUo.010.011.000000012";
		Identifier identifier = new Identifier(pdrID);
		assertEquals(10, identifier.getInstanceNumber());
		assertEquals(11, identifier.getProjectNumber());
		assertEquals(12, identifier.getObjectNumber());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringSchemaAspect() throws Exception
	{
		String pdrID = "pdrAo.013.014.000000015";
		Identifier identifier = new Identifier(pdrID);
		assertEquals("http://pdr.bbaw.de/schema/aodl.xsd", identifier.getSchemaURL());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsToString() throws Exception
	{
		String type = "pdrRo";
		String instance = "005";
		String project = "002";
		String object = "000000003";
		Identifier identifer = new Identifier(type, instance, project, object);
		assertEquals(type + "." + instance + "." + project + "." + object, identifer.toString());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsEquals() throws Exception
	{
		String type = "pdrAo";
		String instance = "001";
		String project = "002";
		String object = "000000037";
		Identifier identifier = new Identifier(type, instance, project, object);
		assertTrue(identifier.equals(new Identifier(type + "." + instance + "." + project + "." + object)));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsNotEqual() throws Exception
	{
		String type = "pdrUo";
		String instance = "051";
		String project = "002";
		String object = "000080003";
		Identifier identifier = new Identifier(type, instance, project, object);
		assertFalse(identifier.equals(new Identifier("pdrPo.004.005.000000060")));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsGetType() throws Exception
	{
		String type = "pdrPo";
		String instance = "001";
		String project = "002";
		String object = "000000003";
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals(PDRType.PERSON, identifier.getType());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsGetNumbers() throws Exception
	{
		String type = "pdrUo";
		String instance = "001";
		String project = "052";
		String object = "100000003";
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals(1, identifier.getInstanceNumber());
		assertEquals(52, identifier.getProjectNumber());
		assertEquals(100000003, identifier.getObjectNumber());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorStringsSchemaAspect() throws Exception
	{
		String type = "pdrPo";
		String instance = "001";
		String project = "002";
		String object = "000000003";
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals("http://pdr.bbaw.de/schema/podl.xsd", identifier.getSchemaURL());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesToString() throws Exception
	{
		PDRType type = PDRType.REFERENCE;
		int instance = 24;
		int project = 56;
		int object = 12345676;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals("pdrRo.024.056.012345676", identifier.toString());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesEquals() throws Exception
	{
		PDRType type = PDRType.PERSON;
		int instance = 4;
		int project = 123;
		int object = 45676;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertTrue(identifier.equals(new Identifier("pdrPo.004.123.000045676")));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesNotEqual() throws Exception
	{
		PDRType type = PDRType.REFERENCE;
		int instance = 888;
		int project = 213;
		int object = 1;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertFalse(identifier.equals(new Identifier("pdrRo.213.888.000000001")));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesGetType() throws Exception
	{
		PDRType type = PDRType.ASPECT;
		int instance = 24;
		int project = 56;
		int object = 12345676;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals(PDRType.ASPECT, identifier.getType());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesGetNumbers() throws Exception
	{
		PDRType type = PDRType.USER;
		int instance = 1;
		int project = 1;
		int object = 1;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals(1, identifier.getInstanceNumber());
		assertEquals(1, identifier.getProjectNumber());
		assertEquals(1, identifier.getObjectNumber());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testConstructorValuesSchemaAspect() throws Exception
	{
		PDRType type = PDRType.REFERENCE;
		int instance = 24;
		int project = 56;
		int object = 12345676;
		Identifier identifier = new Identifier(type, instance, project, object);
		assertEquals("http://pdr.bbaw.de/schema/rodl_mods.xsd", identifier.getSchemaURL());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringInvalidType() throws Exception
	{
		new Identifier("pDRPo.001.001.000000001");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringInvalidInstance() throws Exception
	{
		new Identifier("pdrUo.000.001.000000002");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringInvalidProject() throws Exception
	{
		new Identifier("pdrUo.003.04.000000032");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringInvalidObject() throws Exception
	{
		new Identifier("pdrRo.003.001.00000000x");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringsInvalidType() throws Exception
	{
		new Identifier("", "001", "002", "000066666");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringsInvalidInstance() throws Exception
	{
		new Identifier("pdrAo", "0001", "002", "000066666");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringsInvalidProject() throws Exception
	{
		new Identifier("pdrPo", "010", "000", "000066666");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorStringsInvalidObject() throws Exception
	{
		new Identifier("pdrRo", "501", "072", "66666");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorValuesInvalidInstance() throws Exception
	{
		new Identifier(PDRType.REFERENCE, 4444, 23, 333);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorValuesInvalidProject() throws Exception
	{
		new Identifier(PDRType.ASPECT, 21, -1, 333);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = InvalidIdentifierException.class)
	public final void testConstructorValuesInvalidObject() throws Exception
	{
		new Identifier(PDRType.USER, 21, 111, 0);
	}

}
