package org.bbaw.pdr.allies.client.tests;

import static org.junit.Assert.assertEquals;

import java.net.URI;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Vector;

import org.apache.solr.client.solrj.SolrServer;
import org.apache.solr.client.solrj.impl.HttpSolrServer;
import org.bbaw.pdr.allies.client.Configuration;
import org.bbaw.pdr.allies.client.IDRange;
import org.bbaw.pdr.allies.client.Identifier;
import org.bbaw.pdr.allies.client.PDRType;
import org.bbaw.pdr.allies.client.Repository;
import org.bbaw.pdr.allies.client.Utilities;
import org.bbaw.pdr.padre.PadreConfiguration;
import org.bbaw.pdr.padre.client.PadreClient;
import org.bbaw.pdr.solr.SolrConfiguration;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * @author Fabian Körner
 */
public class TestIngestObjects
{

	private static final org.bbaw.pdr.allies.client.tests.Configuration TEST_CONFIG = new ConfigurationFromFile();
	private static PadreClient _padreClient;
	private static SolrServer _solrServer;

	/**
	 * @throws Exception Exception.
	 */
	@BeforeClass
	public static void setUp() throws Exception
	{
		Configuration.getInstance().setAxis2BaseURL(TEST_CONFIG.getAxis2BaseUri());
		Configuration.getInstance().setPDRUser("pdrUo.001.002.000000001", TEST_CONFIG.getPDRUserPW(1, 2, 1));

		PadreConfiguration.getInstance().setCredentials(URI.create(TEST_CONFIG.getPadreUri()), TEST_CONFIG.getPadreUsername(), TEST_CONFIG.getPadrePassword());
		SolrConfiguration.getInstance().setBaseURI(URI.create(TEST_CONFIG.getSolrUri()));
		_solrServer = SolrConfiguration.getInstance().getSolrServer();
		_padreClient = PadreConfiguration.getInstance().getPadreClient();
		_padreClient.createStore("321");
		_padreClient.createStore("322");
		_padreClient.createStore("323");
		_padreClient.createStore("324");
		_padreClient.createStore("325");
	}

	/**
	 * @throws Exception Exception.
	 */
	@AfterClass
	public static void tearDown() throws Exception
	{
		_padreClient.deleteStore("321");
		_padreClient.deleteStore("322");
		_padreClient.deleteStore("323");
		_padreClient.deleteStore("324");
		_padreClient.deleteStore("325");
		_solrServer.deleteByQuery("id.project:801");
		_solrServer.deleteByQuery("id.project:802");
		_solrServer.deleteByQuery("id.project:803");
		_solrServer.deleteByQuery("id.project:804");
		_solrServer.deleteByQuery("id.project:805");
		((HttpSolrServer) _solrServer).shutdown();
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (timeout = 2000)
	public final void testIngestObjectsByRange() throws Exception
	{
		List<IDRange> ranges = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 801, 1, 10);
		assertEquals(0, ranges.size());
		Vector<String> objects = new Vector<String>();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.801.100000011"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.801.100000012"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.801.100000033"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.801.100000044"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.801.100000032"));

		Repository.ingestObjects(1, 801, objects);

		List<IDRange> result = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 801, 1, 10);
		assertEquals(1, result.size());
		assertEquals(1, result.get(0).getLowerBound());
		assertEquals(5, result.get(0).getUpperBound());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (timeout = 2000)
	public final void testIngestObjectsByContent() throws Exception
	{
		List<IDRange> ranges = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 802, 1, 10);
		assertEquals(0, ranges.size());
		Vector<String> objects = new Vector<String>();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.802.100000011"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.802.100000012"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.802.100000033"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.802.100000044"));
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.802.100000032"));

		Repository.ingestObjects(1, 802, objects);

		List<IDRange> result = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 802, 1, 10);
		assertEquals(1, result.size());

		List<String> returnedObjects = Utilities.getObjects(PDRType.PERSON, 1, 802, result.get(0).getLowerBound(), result.get(0).getUpperBound());
		assertEquals(5, returnedObjects.size());
		assertEquals(returnedObjects.get(0), TestIngestObjects.createNewPersonObject("pdrPo.001.802.000000001"));
		assertEquals(returnedObjects.get(1), TestIngestObjects.createNewPersonObject("pdrPo.001.802.000000002"));
		assertEquals(returnedObjects.get(2), TestIngestObjects.createNewPersonObject("pdrPo.001.802.000000003"));
		assertEquals(returnedObjects.get(3), TestIngestObjects.createNewPersonObject("pdrPo.001.802.000000004"));
		assertEquals(returnedObjects.get(4), TestIngestObjects.createNewPersonObject("pdrPo.001.802.000000005"));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (timeout = 5000)
	public final void testIngestSingleObjectsByRange() throws Exception
	{
		List<IDRange> ranges = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 803, 1, 10);
		assertEquals(0, ranges.size());

		Vector<String> objects = new Vector<String>();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.803.100000011"));
		Repository.ingestObjects(1, 803, objects);

		objects.removeAllElements();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.803.100000012"));
		Repository.ingestObjects(1, 803, objects);

		objects.removeAllElements();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.803.100000033"));
		Repository.ingestObjects(1, 803, objects);

		objects.removeAllElements();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.803.100000044"));
		Repository.ingestObjects(1, 803, objects);

		objects.removeAllElements();
		objects.add(TestIngestObjects.createNewPersonObject("pdrPo.001.803.100000032"));
		Repository.ingestObjects(1, 803, objects);

		List<IDRange> result = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 803, 1, 10);
		assertEquals(1, result.size());
		assertEquals(1, result.get(0).getLowerBound());
		assertEquals(5, result.get(0).getUpperBound());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (timeout = 2000)
	public final void testReturnedMapOfIngestObjectsByIdentifiers() throws Exception
	{
		List<IDRange> ranges = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 804, 1, 10);
		assertEquals(0, ranges.size());

		String[] temporaryIDStrings = {"pdrPo.001.804.100000033", "pdrPo.001.804.100000011", "pdrPo.001.804.100000012", "pdrPo.001.804.100000032",
										"pdrPo.001.804.100000044", };

		String[] persistentIDStrings = {"pdrPo.001.804.000000001", "pdrPo.001.804.000000002", "pdrPo.001.804.000000003", "pdrPo.001.804.000000004",
										 "pdrPo.001.804.000000005", };

		Vector<String> objects = new Vector<String>();
		for (String id : temporaryIDStrings)
		{
			objects.add(TestIngestObjects.createNewPersonObject(id));
		}

		Map<Identifier, Identifier> returnedMap = Repository.ingestObjects(1, 804, objects);
		assertEquals(5, returnedMap.size());

		Identifier key, value;
		for (Entry<Identifier, Identifier> entry : returnedMap.entrySet())
		{
			key = entry.getKey();
			value = entry.getValue();
			if (key.equals(new Identifier(temporaryIDStrings[0])))
				{ assertEquals(value, new Identifier(persistentIDStrings[0])); }
			if (key.equals(new Identifier(temporaryIDStrings[1])))
				{ assertEquals(value, new Identifier(persistentIDStrings[1])); }
			if (key.equals(new Identifier(temporaryIDStrings[2])))
				{ assertEquals(value, new Identifier(persistentIDStrings[2])); }
			if (key.equals(new Identifier(temporaryIDStrings[3])))
				{ assertEquals(value, new Identifier(persistentIDStrings[3])); }
			if (key.equals(new Identifier(temporaryIDStrings[4])))
				{ assertEquals(value, new Identifier(persistentIDStrings[4])); }
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (timeout = 2000)
	public final void testReturnedMapOfIngestObjectsByString() throws Exception
	{

		List<IDRange> ranges = Utilities.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 805, 1, 10);
		assertEquals(0, ranges.size());

		String[] temporaryIDStrings = {
				"pdrPo.001.805.100000033", "pdrPo.001.805.100000011", "pdrPo.001.805.100000012", "pdrPo.001.805.100000032", "pdrPo.001.805.100000044",
		};

		String[] persistentIDStrings = {
				"pdrPo.001.805.000000001", "pdrPo.001.805.000000002", "pdrPo.001.805.000000003", "pdrPo.001.805.000000004", "pdrPo.001.805.000000005",
		};

		Vector<String> objects = new Vector<String>();
		for (String id : temporaryIDStrings)
		{
			objects.add(TestIngestObjects.createNewPersonObject(id));
		}

		Map<Identifier, Identifier> returnedMap = Repository.ingestObjects(1, 805, objects);
		assertEquals(5, returnedMap.size());

		String key, value;
		for (Entry<Identifier, Identifier> entry : returnedMap.entrySet())
		{
			key = entry.getKey().toString();
			value = entry.getValue().toString();
			if (key.equals(temporaryIDStrings[0]))
				{ assertEquals(value, persistentIDStrings[0]); }
			if (key.equals(temporaryIDStrings[1]))
				{ assertEquals(value, persistentIDStrings[1]); }
			if (key.equals(temporaryIDStrings[2]))
				{ assertEquals(value, persistentIDStrings[2]); }
			if (key.equals(temporaryIDStrings[3]))
				{ assertEquals(value, persistentIDStrings[3]); }
			if (key.equals(temporaryIDStrings[4]))
				{ assertEquals(value, persistentIDStrings[4]); }
		}
	}

	private static String createNewPersonObject(final String id)
	{
		return new String("<person xmlns=\"http://pdr.bbaw.de/namespaces/podl/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""
				+ " xsi:schemaLocation=\"http://pdr.bbaw.de/namespaces/podl/ http://pdr.bbaw.de/schema/podl.xsd\" id=\"" + id + "\">"
				+ "<record><revision ref=\"0\" timestamp=\"2012-11-30T04:48:08\" authority=\"pdrUo.002.762.000000002\"/></record></person>");
	}

}
