/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import java.util.List;

/**
 * A collection of methods for creation of contents of SOAP data parts
 * of HTTP requests.
 * @author Fabian Körner
 */
public final class DataParts
{

	private static final String XML_DECLARATION = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	private static final String ENVELOPE_AND_BODY_OPEN =
		"<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\"><soapenv:Body>";
	private static final String ENVELOPE_AND_BODY_CLOSE = "</soapenv:Body></soapenv:Envelope>";

	private DataParts()
	{
	}

	/**
	 * Creates SOAP request document for method: getVersion.
	 * @return SOAP request document.
	 */
	public static String createGetVersion()
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getVersion xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\" />");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getTime.
	 * @return SOAP request document.
	 */
	public static String createGetTime()
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getTime xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\" />");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getOccupiedObjectIDRanges.
	 * @param type PDR type.
	 * @param instance PDR instance ID.
	 * @param project PDR project ID.
	 * @param min Lower bound.
	 * @param max Upper bound.
	 * @return SOAP request document.
	 */
	public static String createGetOccupiedObjectIDRanges(final PDRType type, final int instance, final int project, final int min, final int max)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getOccupiedIDRanges xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Max", Integer.toString(max)));
		sb.append(DataParts.fillParameterElement("Min", Integer.toString(min)));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append(DataParts.fillParameterElement("Instance", Integer.toString(instance)));
		sb.append(DataParts.fillParameterElement("Type", type.toString()));
		sb.append("</axis2ns3:getOccupiedIDRanges>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: setCategories.
	 * @param provider Categories provider.
	 * @param categories Categories as a <code>java.lang.String</code>.
	 * @return SOAP request document.
	 */
	public static String createSetCategories(final String provider, final String categories)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:setCategories xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Provider", provider));
		sb.append(DataParts.fillParameterElement("Categories", categories));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:setCategories>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getCategories.
	 * @param provider Categories provider.
	 * @return SOAP request document.
	 */
	public static String createGetCategories(final String provider)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getCategories xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Provider", provider));
		sb.append("</axis2ns3:getCategories>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getCategoryProviders.
	 * @return SOAP request document.
	 */
	public static String createGetCategoryProviders()
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getCategoryProviders xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\" />");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getObjects.
	 * @param type PDR type.
	 * @param instance PDR instance ID.
	 * @param project PDR project ID.
	 * @param min Lower bound.
	 * @param max Upper bound.
	 * @return SOAP request document.
	 */
	public static String createGetObjects(final PDRType type, final int instance, final int project, final int min, final int max)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getObjects xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Max", Integer.toString(max)));
		sb.append(DataParts.fillParameterElement("Min", Integer.toString(min)));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append(DataParts.fillParameterElement("Instance", Integer.toString(instance)));
		sb.append(DataParts.fillParameterElement("Type", type.toString()));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:getObjects>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getProjectSizes.
	 * @return SOAP request document.
	 */
	public static String createGetProjectSizes()
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getProjectSizes xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\" />");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: removeCategories.
	 * @param provider Categories provider.
	 * @return SOAP request document.
	 */
	public static String createRemoveCategories(final String provider)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:removeCategories xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Provider", provider));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:removeCategories>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getModifiedObjects.
	 * @param instance PDR instance ID.
	 * @param project PDR project ID.
	 * @param timestamp Timestamp.
	 * @return SOAP request document.
	 */
	public static String createGetModifiedObjects(final int instance, final int project, final String timestamp)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getModifiedObjects xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("Timestamp", timestamp));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append(DataParts.fillParameterElement("Instance", Integer.toString(instance)));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:getModifiedObjects>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: ingestObjects.
	 * @param instance PDR instance ID.
	 * @param project PDR project ID.
	 * @param objects PDR objects.
	 * @return SOAP request document.
	 */
	public static String createIngestObjects(final int instance, final int project, 	final List<String> objects)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:ingestObjects xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.createObjectElements(objects));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append(DataParts.fillParameterElement("Instance", Integer.toString(instance)));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:ingestObjects>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: modifyObjects.
	 * @param instance PDR instance ID.
	 * @param project PDR project ID.
	 * @param objects PDR objects.
	 * @param force Overwrite conflicting objects.
	 * @return SOAP request document.
	 */
	public static String createModifyObjects(final int instance, final int project, final List<String> objects, final boolean force)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:modifyObjects xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.createObjectElements(objects));
		sb.append(DataParts.fillParameterElement("Force", Boolean.toString(force)));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append(DataParts.fillParameterElement("Instance", Integer.toString(instance)));
		sb.append(DataParts.fillParameterElement("Credentials", Configuration.getInstance().getCredentials()));
		sb.append("</axis2ns3:modifyObjects>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	/**
	 * Creates SOAP request document for method: getUserID.
	 * @param username User's name.
	 * @param project PDR project ID.
	 * @return SOAP request document.
	 */
	public static String createGetUserID(final String username, final int project)
	{
		StringBuilder sb = new StringBuilder(XML_DECLARATION);
		sb.append(ENVELOPE_AND_BODY_OPEN);
		sb.append("<axis2ns3:getUserID xmlns:axis2ns3=\"http://allies.pdr.bbaw.org\">");
		sb.append(DataParts.fillParameterElement("User", username));
		sb.append(DataParts.fillParameterElement("Project", Integer.toString(project)));
		sb.append("</axis2ns3:getUserID>");
		sb.append(ENVELOPE_AND_BODY_CLOSE);
		return sb.toString();
	}

	private static String fillParameterElement(final String name, final String value)
	{
		StringBuilder sb = new StringBuilder();
		sb.append('<');
		sb.append(name);
		sb.append('>');
		sb.append(value);
		sb.append('<');
		sb.append('/');
		sb.append(name);
		sb.append('>');
		return sb.toString();
	}

	private static String createObjectElements(final List<String> objects)
	{
		StringBuilder sb = new StringBuilder();
		int count = 0;
		String name;
		String value;
		for (String object : objects)
		{
			name = "Object" + Integer.toString(count++);
			value = Tools.cloak(object);
			sb.append(DataParts.fillParameterElement(name, value));
		}
		return sb.toString();
	}
}
