/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.bbaw.pdr.allies.error.InvalidIdentifierException;

/**
 * Represent a PDR ID and its parts.
 * @author Fabian Körner
 */
public class Identifier
{

	private static final Pattern PATTERN_ID = Pattern.compile("^(pdr[APRU]o)\\.(\\d{3})\\.(\\d{3})\\.(\\d{9})$");
	private static final int GROUP_TYPE = 1;
	private static final int GROUP_INSTANCE = 2;
	private static final int GROUP_PROJECT = 3;
	private static final int GROUP_OBJECT = 4;

	private static final int LENGTH_ID_INSTANCE = 3;
	private static final int LENGTH_ID_PROJECT = 3;
	private static final int LENGTH_ID_OBJECT = 9;

	/** Schema URL for aspects. */
	public static final String SCHEMA_URL_ASPECT = "http://pdr.bbaw.de/schema/aodl.xsd";
	/** Schema URL for persons. */
	public static final String SCHEMA_URL_PERSON = "http://pdr.bbaw.de/schema/podl.xsd";
	/** Schema URL for users. */
	public static final String SCHEMA_URL_USER = "http://pdr.bbaw.de/schema/uodl.xsd";
	/** Schema URL for references. */
	public static final String SCHEMA_URL_REFERENCE = "http://pdr.bbaw.de/schema/rodl_mods.xsd";

	private String _id;
	private PDRType _type;
	private int _instance;
	private int _project;
	private int _object;

	/**
	 * Constructor with PDR ID as String.
	 * @param id PDR ID.
	 * @throws InvalidIdentifierException ID does not match pattern.
	 */
	public Identifier(final String id) throws InvalidIdentifierException
	{
		_id = id;
		this.parseIDString();
		this.validateNumberParts();
	}

	/**
	 * Constructor with single parts of the ID as Strings.
	 * @param type PDR object type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param object Object ID.
	 * @throws InvalidIdentifierException Combined ID does not match pattern.
	 */
	public Identifier(final String type, final String instance, final String project, final String object) throws InvalidIdentifierException
	{
		this(Identifier.getIDFromSingleStrings(type, instance, project, object));
	}

	/**
	 * Constructor with single parts of the ID.
	 * @param type PDR object type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param object Object ID.
	 * @throws InvalidIdentifierException Combined ID does not match pattern.
	 */
	public Identifier(final PDRType type, final int instance, final int project, final int object) 	throws InvalidIdentifierException
	{
		this(Identifier.getIDFromSingleParts(type, instance, project, object));
	}

	private void parseIDString() throws InvalidIdentifierException
	{
		Matcher m = PATTERN_ID.matcher(_id);
		if (!m.find())
		{
			throw new InvalidIdentifierException(_id);
		}
		else
		{
			_type = typeFromString(m.group(GROUP_TYPE));
			_instance = Integer.parseInt(m.group(GROUP_INSTANCE));
			_project = Integer.parseInt(m.group(GROUP_PROJECT));
			_object = Integer.parseInt(m.group(GROUP_OBJECT));
		}
	}

	private void validateNumberParts() throws InvalidIdentifierException
	{
		if (_instance < 1 || _project < 1 || _object < 1)
		{
			throw new InvalidIdentifierException("number parts must be greater than zero: " + _id);
		}
	}

	private static String getIDFromSingleStrings(final String type, final String instance, final String project, final String object)
	{
		StringBuilder sb = new StringBuilder();
		sb.append(type);
		sb.append('.');
		sb.append(instance);
		sb.append('.');
		sb.append(project);
		sb.append('.');
		sb.append(object);
		return sb.toString();
	}

	private static String getIDFromSingleParts(final PDRType type, final int instance, final int project, final int object)
	{
		StringBuilder sb = new StringBuilder();
		sb.append(type.toString());
		sb.append('.');
		sb.append(Identifier.addZerosInFront(Integer.toString(instance), LENGTH_ID_INSTANCE));
		sb.append('.');
		sb.append(Identifier.addZerosInFront(Integer.toString(project), LENGTH_ID_PROJECT));
		sb.append('.');
		sb.append(Identifier.addZerosInFront(Integer.toString(object), LENGTH_ID_OBJECT));
		return sb.toString();
	}

	private static String addZerosInFront(final String value, final int length)
	{
		StringBuilder sb = new StringBuilder(value);
		while (sb.length() < length)
		{
			sb.insert(0, '0');
		}
		return sb.toString();
	}

	/**
	 * @see java.lang.Object#toString()
	 * @return Object's string representation.
	 */
	@Override
	public final String toString()
	{
		return _id;
	}

	/**
	 * @see java.lang.Object#equals(Object)
	 * @param obj Other object.
	 * @return Objects are equal.
	 */
	@Override
	public final boolean equals(final Object obj)
	{
		if (obj instanceof Identifier)
		{
			return _id.equals(((Identifier) obj).toString());
		}
		else
		{
			return false;
		}
	}

	/**
	 * @see java.lang.Object#hashCode()
	 * @return Object's hash code.
	 */
	@Override
	public final int hashCode()
	{
		return _id.hashCode();
	}

	/**
	 * Returns the PDR type.
	 * @return PDR type.
	 */
	public final PDRType getType()
	{
		return _type;
	}

	/**
	 * Returns the instance ID.
	 * @return Instance ID.
	 */
	public final int getInstanceNumber()
	{
		return _instance;
	}

	/**
	 * Returns the project ID.
	 * @return Project ID.
	 */
	public final int getProjectNumber()
	{
		return _project;
	}

	/**
	 * Returns the object ID.
	 * @return Object ID.
	 */
	public final int getObjectNumber()
	{
		return _object;
	}

	/**
	 * Translates the given prefix string to the corresponding PDR type.
	 * @param prefix Prefix.
	 * @return PDR type.
	 */
	public static PDRType typeFromString(final String prefix)
	{
		if (prefix.equals(PDRType.ASPECT.toString()))
		{
			return PDRType.ASPECT;
		}
		else if (prefix.equals(PDRType.PERSON.toString()))
		{
			return PDRType.PERSON;
		}
		else if (prefix.equals(PDRType.REFERENCE.toString()))
		{
			return PDRType.REFERENCE;
		}
		else if (prefix.equals(PDRType.USER.toString()))
		{
			return PDRType.USER;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Returns the XML schema URL according to identifier's type.
	 * @return XML schema URL.
	 */
	public final String getSchemaURL()
	{
		switch (_type)
		{
		case ASPECT:
			return SCHEMA_URL_ASPECT;
		case PERSON:
			return SCHEMA_URL_PERSON;
		case USER:
			return SCHEMA_URL_USER;
		case REFERENCE:
			return SCHEMA_URL_REFERENCE;
		default:
			throw new IllegalArgumentException(_type.toString());
		}
	}

	/**
	 * At risk of getting a <code>null</code> reference, this method can
	 * be used to create an instance of this class when the validity of
	 * the given <code>idString</code> is secured during runtime by another
	 * mechanim.
	 * @param idString PDR ID in its String form.
	 * @return <code>null</code> or an instance of this class.
	 */
	public static Identifier create(final String idString)
	{
		Identifier id = null;
		try
		{
			id =  new Identifier(idString);
		}
		catch (InvalidIdentifierException e)
		{
			/* unhandled, see javadoc comment */
		}
		return id;
	}
}
