/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-Allies.
 *
 *	PDR-Allies is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-Allies is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-Allies.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.allies.client;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMElement;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.impl.client.CloseableHttpClient;
import org.bbaw.pdr.allies.AlliesConstants;
import org.bbaw.pdr.allies.error.InvalidIdentifierException;
import org.bbaw.pdr.allies.error.PDRAlliesClientException;

/**
 * Encapsulates methods for the access to the PDR repository.
 * @author Fabian Körner
 */
public final class Repository
{

	private Repository()
	{
	}

	/**
	 * Returns objects which were modified since a given timestamp (format: <code>0000-00-00T00:00:00</code>).
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param timestamp Timestamp.
	 * @return List of modified objects (XML).
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 */
	public static List<String> getModifiedObjects(final int instance, final int project, final String timestamp) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			List<String> result = new Vector<String>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/getModifiedObjects";
			String dataPart = DataParts.createGetModifiedObjects(instance, project, timestamp);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);

			OMElement objectsElement = alliesResponseElement.getFirstElement();

			Iterator<?> objectsIterator = objectsElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_OBJECT,
							AlliesConstants.QNAME_PREFIX_ALLIES));

			OMElement objectElement;
			while (objectsIterator.hasNext())
			{
				objectElement = (OMElement) objectsIterator.next();
				result.add(Tools.uncloak(objectElement.getText()));
			}

			return  result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Returns a map of available project IDs to the corresponding object
	 * counts.
	 * @return Map of project IDs to object counts.
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 */
	public static Map<Integer, Integer> getProjectSizes() throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			final Map<Integer, Integer> resultMap = new HashMap<Integer, Integer>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/getProjectSizes";
			String dataPart = DataParts.createGetProjectSizes();

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement mapElement = alliesResponseElement.getFirstElement();
			Iterator<?> mapIterator = mapElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_KEY_VALUE_PAIR,
							AlliesConstants.QNAME_PREFIX_ALLIES));

			OMElement pairElement;
			OMElement keyElement;
			OMElement valueElement;
			while (mapIterator.hasNext())
			{
				pairElement = (OMElement) mapIterator.next();
				keyElement = pairElement.getFirstChildWithName(
						new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
								AlliesConstants.OMCHILD_LNAME_KEY, AlliesConstants.QNAME_PREFIX_ALLIES));
				valueElement = pairElement.getFirstChildWithName(
						new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
								AlliesConstants.OMCHILD_LNAME_VALUE, AlliesConstants.QNAME_PREFIX_ALLIES));
				resultMap.put(
						new Integer(Tools.uncloak(keyElement.getText())),
						new Integer(Tools.uncloak(valueElement.getText())));
			}

			return resultMap;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Returns PDR user ID connected to provided username and project.
	 * @param username Username.
	 * @param project User's project number.
	 * @return User's PDR ID.
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 */
	public static String getUserID(final String username, final int project) throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			String result = null;

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/getUserID";
			String dataPart = DataParts.createGetUserID(username, project);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement idElement = alliesResponseElement.getFirstElement();

			result = idElement.getText();

			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Returns current PDR server time.
	 * @return User's PDR ID.
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 */
	public static String getTime() throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			String result = null;

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/getTime";
			String dataPart = DataParts.createGetTime();

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement timeElement = alliesResponseElement.getFirstElement();

			result = timeElement.getText();

			return result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Ingests the provided new objects into the repository and returns a
	 * mapping of former temporary to then persistent IDs.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param objects PDR objects to be ingested.
	 * @return Map of temporary to persistent IDs.
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 * @throws InvalidIdentifierException Invalid PDR ID provided.
	 */
	public static Map<Identifier, Identifier> ingestObjects(final int instance, final int project, final List<String> objects)
			throws PDRAlliesClientException, InvalidIdentifierException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			Map<Identifier, Identifier> resultMap = new HashMap<Identifier, Identifier>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/ingestObjects";
			String dataPart = DataParts.createIngestObjects(instance, project, objects);

			httpClient = Tools.buildHttpClient();

			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);
			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement mapElement = alliesResponseElement.getFirstElement();

			Iterator<?> mapIterator = mapElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_KEY_VALUE_PAIR,
							AlliesConstants.QNAME_PREFIX_ALLIES));
			OMElement pairElement;
			OMElement keyElement;
			OMElement valueElement;
			while (mapIterator.hasNext())
			{
				pairElement = (OMElement) mapIterator.next();
				keyElement = pairElement.getFirstChildWithName(
						new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
								AlliesConstants.OMCHILD_LNAME_KEY, AlliesConstants.QNAME_PREFIX_ALLIES));
				valueElement = pairElement.getFirstChildWithName(
						new QName(AlliesConstants.QNAME_NS_URI_ALLIES,
								AlliesConstants.OMCHILD_LNAME_VALUE, AlliesConstants.QNAME_PREFIX_ALLIES));
				resultMap.put(
						new Identifier(Tools.uncloak(keyElement.getText())),
						new Identifier(Tools.uncloak(valueElement.getText())));
			}

			return resultMap;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}

	/**
	 * Writes modified objects to the repository.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param objects Modified objects.
	 * @param force Enforce modification (e.g. for solved conflicts).
	 * @return List of conflicting objects (server version).
	 * @throws PDRAlliesClientException Error during processing the HTTP request.
	 */
	public static List<String> modifyObjects(final int instance, final int project, final List<String> objects, final boolean force)
			throws PDRAlliesClientException
	{
		CloseableHttpClient httpClient = null;
		try
		{
			final Vector<String> result = new Vector<String>();

			String methodURL = Configuration.getInstance().getAxis2BaseURL() + "/services/Repository/modifyObjects";
			String dataPart = DataParts.createModifyObjects(instance, project, objects, force);

			httpClient = Tools.buildHttpClient();
			HttpUriRequest post = Tools.createHttpPostRequest(methodURL, dataPart);

			HttpResponse httpResponse = httpClient.execute(post);

			OMElement alliesResponseElement = Tools.getAlliesResponseElement(httpResponse);
			OMElement objectsElement = alliesResponseElement.getFirstElement();

			Iterator<?> objectsIterator = objectsElement.getChildrenWithName(
					new QName(AlliesConstants.QNAME_NS_URI_ALLIES, AlliesConstants.OMCHILD_LNAME_OBJECT,
							AlliesConstants.QNAME_PREFIX_ALLIES));
			OMElement objectElement;
			while (objectsIterator.hasNext())
			{
				objectElement = (OMElement) objectsIterator.next();
				result.add(Tools.uncloak(objectElement.getText()));
			}

			return  result;
		}
		catch (IOException cause)
		{
			throw new PDRAlliesClientException(cause);
		}
		finally
		{
			Tools.closeHttpClient(httpClient);
		}
	}
}
