/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.padre;

import java.util.List;
import java.util.Vector;

import org.bbaw.pdr.padre.request.PadreID;
import org.bbaw.pdr.utils.IDRange;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRType;
import org.bbaw.pdr.utils.ValidationException;

/**
 * Returns a list of ID ranges in a certain project.
 * @author Fabian Körner
 */
public final class PadreGetIDRanges extends AbstractPadreCommand<List<IDRange>>
{

	private PDRType _type;
	private int _instance;
	private int _project;
	private int _lowerBound;
	private int _upperBound;

	/**
	 * Returns ID ranges of available objects within given bounds.
	 * @param type Type of objects.
	 * @param instance Instance ID of objects.
	 * @param project Project ID of objects.
	 * @param lowerBound Lower bound of range.
	 * @param upperBound Upper bound of range.
	 * @throws PadreCommandException Could not get ID ranges.
	 */
	public PadreGetIDRanges(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound) throws PadreCommandException
	{
		super();
		_type = type;
		_instance = instance;
		_project = project;
		_lowerBound = lowerBound;
		_upperBound = upperBound;
	}

	/**
	 * @see org.bbaw.pdr.padre.AbstractPadreCommand#execute()
	 * @return ID ranges.
	 * @throws PadreCommandException Could not successfully execute PaDRe command.
	 */
	@Override
	public List<IDRange> execute() 	throws PadreCommandException
	{
		try
		{
			String padreStoreID = Identifier.intToHexString(_project, 3);
			String padreType = _type.toString();

			List<PadreID> padreIDs = super.getPadreClient().ranges(padreStoreID, padreType);
			List<IDRange> ranges = this.padreIDsToIDRanges(padreIDs);
			return IDRange.applyLimitsToRanges(ranges, _lowerBound, _upperBound);

		}
		catch (Exception cause)
		{
			throw new PadreCommandException(cause);
		}
	}

	private List<IDRange> padreIDsToIDRanges(final List<PadreID> padreIDs) throws ValidationException
	{
		List<IDRange> result = new Vector<IDRange>();
		for (PadreID padreID : padreIDs)
		{
			int lowerBound = IDRange.safeLongToInt(padreID.getLowerBound().getValueObject());
			int upperBound = IDRange.safeLongToInt(padreID.getUpperBound().getValueObject());
			result.add(new IDRange(_type, _instance, _project, lowerBound, upperBound));
		}
		return result;
	}

}
