/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.padre;

import java.util.List;
import java.util.Vector;

import org.apache.commons.codec.binary.Base64;
import org.bbaw.pdr.padre.request.PadreID;
import org.bbaw.pdr.padre.request.PadreObject;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRObject;

/**
 * Reads a span of XML formatted PDR objects from PaDRe.
 * @author Fabian Körner
 */
public final class PadreSpanReadXml extends AbstractPadreCommand<List<PDRObject>>
{

	private Identifier _lowerBound;
	private Identifier _upperBound;

	/**
	 * Reads all objects within the provided span.
	 * @param minIdentifier Lower bound.
	 * @param maxIdentifier Upper bound.
	 * @throws PadreCommandException Could not read from PaDRe.
	 */
	public PadreSpanReadXml(final String minIdentifier, final String maxIdentifier) throws PadreCommandException
	{
		super();
		try
		{
			_lowerBound = new Identifier(minIdentifier);
			_upperBound = new Identifier(maxIdentifier);
		}
		catch (Exception cause)
		{
			throw new PadreCommandException(cause);
		}
	}

	/**
	 * @see org.bbaw.pdr.padre.AbstractPadreCommand#execute()
	 * @return PDR objects.
	 * @throws PadreCommandException Could not successfully execute PaDRe command.
	 */
	@Override
	public List<PDRObject> execute() throws PadreCommandException
	{
		List<PDRObject> result = new Vector<PDRObject>();
		try
		{
			PadreID lower = _lowerBound.toPadreID();
			PadreID upper = _upperBound.toPadreID();
			PadreID span = new PadreID(lower.toString() + "-" + upper.toString());
			List<PadreID> padreIDs = new Vector<PadreID>();
			padreIDs.add(span);
			List<PadreObject> padreObjects = super.getPadreClient().read(padreIDs);
			for (PadreObject padreObject : padreObjects)
			{
				String payload = padreObject.getPayload();
				byte[] content = Base64.decodeBase64(payload);
				result.add(new PDRObject(new String(content, "UTF-8")));
			}
		}
		catch (Exception cause)
		{
			throw new PadreCommandException(cause);
		}
		return result;
	}
}
