/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.utils;

import java.util.List;
import java.util.Vector;

/**
 * Range of IDs of a certain PDR type.
 * @author Fabian Körner
 */
public class IDRange
{
	/** Minimal valid value of an instance ID. */
	public static final int MIN_INSTANCE_ID = 1;
	/** Maximal valid value of an instance ID. */
	public static final int MAX_INSTANCE_ID = 999;

	/** Minimal valid value of a project ID. */
	public static final int MIN_PROJECT_ID = 1;
	/** Maximal valid value of a project ID. */
	public static final int MAX_PROJECT_ID = 999;

	/** Minimal valid value of an object ID. */
	public static final int MIN_OBJECT_ID = 1;
	/** Maximal valid value of an object ID. */
	public static final int MAX_OBJECT_ID = 999999999;

	private PDRType _type;
	private int _instance;
	private int _project;
	private int _lowerBound;
	private int _upperBound;

	/**
	 * Range of identifiers.
	 * @param type Type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound.
	 * @param upperBound Upper bound.
	 * @throws ValidationException Invalid bounds.
	 */
	public IDRange(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound) throws ValidationException
	{
		_type = type;
		_instance = instance;
		_project = project;
		_lowerBound = lowerBound;
		_upperBound = upperBound;

		this.checkLimits();
	}

	private void checkLimits() throws ValidationException
	{
		if (_instance < 1 || _instance > 999)
		{
			throw new ValidationException("instance IDs are only valid in range: 1..999; here: " + Integer.toString(_instance));
		}
		if (_project < 1 || _project > 999)
		{
			throw new ValidationException("project IDs are only valid in range: 1..999; here: " + Integer.toString(_project));
		}
		if (_lowerBound < 1 || _lowerBound > 199999999)
		{
			throw new ValidationException("lower bound is only valid in range: 1..199999999; here: " + Integer.toString(_lowerBound));
		}
		if (_upperBound < 1 || _upperBound > 199999999)
		{
			throw new ValidationException("upper bound is only valid in range: 1..199999999; here: " + Integer.toString(_upperBound));
		}
		if (_lowerBound > _upperBound)
		{
			throw new ValidationException("lower is greater than upper bound: " + Integer.toString(_lowerBound) + " > " + Integer.toString(_upperBound));
		}
	}

	/**
	 * Returns the type of objects in the range.
	 * @return Type.
	 */
	public final PDRType getType()
	{
		return _type;
	}

	/**
	 * Returns the instance ID of the range.
	 * @return Instance ID.
	 */
	public final int getInstance()
	{
		return _instance;
	}

	/**
	 * Returns the project ID of the range.
	 * @return Project ID.
	 */
	public final int getProject()
	{
		return _project;
	}

	/**
	 * Returns the lower limit of the range.
	 * @return Lower bound.
	 */
	public final int getLowerBound()
	{
		return _lowerBound;
	}

	/**
	 * Returns the upper limit of the range.
	 * @return Upper bound.
	 */
	public final int getUpperBound()
	{
		return _upperBound;
	}

	/**
	 * Returns the number of objects in the range.
	 * @return Number of objects.
	 */
	public final int count()
	{
		return 1 + _upperBound - _lowerBound;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public final String toString()
	{
		StringBuilder sb = new StringBuilder();
		sb.append("IDRange[");
		sb.append(_type.toString());
		sb.append(',');
		sb.append(Integer.toString(_instance));
		sb.append(',');
		sb.append(Integer.toString(_project));
		sb.append(',');
		sb.append(Integer.toString(_lowerBound));
		sb.append('-');
		sb.append(Integer.toString(_upperBound));
		sb.append(']');
		return sb.toString();
	}

	/**
	 * Casts the provided Integer to Long, if possible.
	 * @param value Integer to be cast.
	 * @return Long cast from Integer.
	 */
	public static int safeLongToInt(final long value)
	{
		if (value < Integer.MIN_VALUE || value > Integer.MAX_VALUE)
		{
			throw new IllegalArgumentException(value + "cannot safely cast " + Long.toString(value) + " to int");
		}
		return (int) value;
	}

	/**
	 * Returns a subset of ranges which fit the provided bounds. Overlapping ranges will be trimmed to fit the bounds.
	 * @param ranges Starting list of ranges.
	 * @param lowerBound Lower bound.
	 * @param upperBound Upper bound.
	 * @return Ranges within the provided bounds.
	 * @throws ValidationException Invalid resulting bounds.
	 */
	public static List<IDRange> applyLimitsToRanges(final List<IDRange> ranges, final int lowerBound, final int upperBound) throws ValidationException
	{
		List<IDRange> result = new Vector<IDRange>();
		for (IDRange range : ranges)
		{
			PDRType type = range.getType();
			int instance = range.getInstance();
			int project = range.getProject();
			int lower = range.getLowerBound();
			int upper = range.getUpperBound();

			if (upper < lowerBound || lower > upperBound)
			{
				/* range out of limits -> ignore */
			}
			else if (lower < lowerBound && upper > upperBound)
			{
				result.add(new IDRange(type, instance, project, lowerBound, upperBound));
			}
			else if (lower < lowerBound && upper >= lowerBound)
			{
				result.add(new IDRange(type, instance, project, lowerBound, upper));
			}
			else if (upper > upperBound && lower <= upperBound)
			{
				result.add(new IDRange(type, instance, project, lower, upperBound));
			}
			else
			{
				result.add(range);
			}
		}
		return result;
	}
}
