/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.utils;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.bbaw.pdr.padre.request.PadreID;
import org.bbaw.pdr.padre.request.PadreIDException;


/**
 * Represents an internal identifier and its components.
 * @author Fabian Körner
 */
public class Identifier
{

	private static final Pattern PATTERN_ID = Pattern.compile("^(pdr[APRU]o)\\.(\\d{3})\\.(\\d{3})\\.(\\d{9})$");

	private static final int GROUP_ID = 0;
	private static final int GROUP_TYPE = 1;
	private static final int GROUP_INSTANCE = 2;
	private static final int GROUP_PROJECT = 3;
	private static final int GROUP_OBJECT = 4;

	private static final int LENGTH_ID_AFTER_INSTANCE = 9;
	private static final int LENGTH_ID_AFTER_PROJECT = 13;
	private static final int LENGTH_ID_COMPLETE = 23;  // pdrAo.000.000.000000000
	private static final int OFFSET_ID_INSTANCE = 6;
	private static final int OFFSET_ID_PROJECT = 10;
	private static final int OFFSET_ID_OBJECT = 14;

	/** XML Schema URL: Aspects. */
	public static final String SCHEMA_URL_ASPECT = "http://pdr.bbaw.de/schema/aodl.xsd";
	/** XML Schema URL: Persons. */
	public static final String SCHEMA_URL_PERSON = "http://pdr.bbaw.de/schema/podl.xsd";
	/** XML Schema URL: Repository Users. */
	public static final String SCHEMA_URL_USER = "http://pdr.bbaw.de/schema/uodl.xsd";
	/** XML Schema URL: References. */
	public static final String SCHEMA_URL_REFERENCE = "http://pdr.bbaw.de/schema/rodl_mods.xsd";

	private String _id;
	private PDRType _type;
	private int _instance;
	private int _project;
	private int _object;

	/**
	 * Constructor with ID as <code>java.lang.String</code>.
	 * @param id PDR ID in common format.
	 * @throws ValidationException Invalid ID.
	 */
	public Identifier(final String id) throws ValidationException
	{
		Matcher m = PATTERN_ID.matcher(id);
		if (!m.find())
		{
			throw new ValidationException("invalid identifier: " + id);
		}
		else
		{
			/* ganze ID */
			_id = m.group(GROUP_ID);
			/* Typ und IDs */
			_type = typeFromString(m.group(GROUP_TYPE));
			_instance = Integer.parseInt(m.group(GROUP_INSTANCE));
			_project = Integer.parseInt(m.group(GROUP_PROJECT));
			_object = Integer.parseInt(m.group(GROUP_OBJECT));
		}
		this.checkForbiddenValues();
	}

	/**
	 * Constructor with separated components as <code>java.lang.String</code>s.
	 * @param type PDR type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param object Object ID.
	 * @throws ValidationException Resulting ID is not valid.
	 */
	public Identifier(final String type, final String instance, final String project, final String object) throws ValidationException
	{
		StringBuilder sb = new StringBuilder();
		sb.append(type);
		sb.append('.');
		sb.append(instance);
		sb.append('.');
		sb.append(project);
		sb.append('.');
		sb.append(object);

		Matcher m = PATTERN_ID.matcher(sb.toString());
		if (!m.find())
		{
			throw new ValidationException("invalid identifier: " + sb.toString());
		}
		else
		{
			_id = m.group(GROUP_ID); /* complete ID */
			_type = typeFromString(m.group(GROUP_TYPE));
			_instance = Integer.parseInt(m.group(GROUP_INSTANCE));
			_project = Integer.parseInt(m.group(GROUP_PROJECT));
			_object = Integer.parseInt(m.group(GROUP_OBJECT));
		}
		this.checkForbiddenValues();
	}

	/**
	 * Constructor with separated components.
	 * @param type PDR type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param object Object ID.
	 * @throws ValidationException Resulting ID is not valid.
	 */
	public Identifier(final PDRType type, final int instance, final int project, final int object) 	throws ValidationException
	{
		StringBuilder sb = new StringBuilder();
		sb.append(type.toString());
		sb.append('.');
		sb.append(Integer.toString(instance));
		while (sb.length() < LENGTH_ID_AFTER_INSTANCE)
		{
			sb.insert(OFFSET_ID_INSTANCE, '0');
		}
		sb.append('.');
		sb.append(Integer.toString(project));
		while (sb.length() < LENGTH_ID_AFTER_PROJECT)
		{
			sb.insert(OFFSET_ID_PROJECT, '0');
		}
		sb.append('.');
		sb.append(Integer.toString(object));
		while (sb.length() < LENGTH_ID_COMPLETE)
		{
			sb.insert(OFFSET_ID_OBJECT, '0');
		}

		Matcher m = PATTERN_ID.matcher(sb.toString());
		if (!m.find())
		{
			throw new ValidationException("invalid identifier: " + sb.toString());
		}
		else
		{
			/* ganze ID */
			_id = m.group(GROUP_ID);
			/* Typ und IDs */
			_type = typeFromString(m.group(GROUP_TYPE));
			_instance = Integer.parseInt(m.group(GROUP_INSTANCE));
			_project = Integer.parseInt(m.group(GROUP_PROJECT));
			_object = Integer.parseInt(m.group(GROUP_OBJECT));
		}
		this.checkForbiddenValues();
	}

	private void checkForbiddenValues() throws ValidationException
	{
		if (0 == _instance)
		{
			throw new ValidationException("instance part of PDR ID must not be zero: " + _instance);
		}
		else if (0 == _project)
		{
			throw new ValidationException("project part of PDR ID must not be zero: " + _project);
		}
		else if (0 == _object)
		{
			throw new ValidationException("object part of PDR ID must not be zero: "  + _object);
		}
		else if (100000000 == _object)
		{
			throw new ValidationException("object part of PDR ID must not be 100000000" + _object);
		}
		else if (199999999 < _object)
		{
			throw new ValidationException("object part of PDR ID must not be greater than 199999999: " + _object);
		}
		else
		{
			return;
		}
	}

	/**
	 * @see java.lang.String#toString()
	 */
	@Override
	public final String toString()
	{
		return _id;
	}

	/**
	 * @see java.lang.Object#equals(Object)
	 */
	@Override
	public final boolean equals(final Object otherObject)
	{
		return this.hashCode() == otherObject.hashCode();
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public final int hashCode()
	{
		return _id.hashCode();
	}

	/**
	 * Returns the PDR type of the identifier.
	 * @return PDR Type.
	 */
	public final PDRType getType()
	{
		return _type;
	}

	/**
	 * Returns the instance ID of the identifier.
	 * @return Instance ID.
	 */
	public final int getInstanceNumber()
	{
		return _instance;
	}

	/**
	 * Returns the project ID of the identifier.
	 * @return Project ID.
	 */
	public final int getProjectNumber()
	{
		return _project;
	}

	/**
	 * Returns the object ID of the identifier.
	 * Liefert die in der ID enthaltene Objektnummer.
	 * @return Object ID.
	 */
	public final int getObjectNumber()
	{
		return _object;
	}

	/**
	 * Converts the provided prefix to a <code>org.bbaw.pdr.utils.PDRType</code>.
	 * @param prefix Prefix.
	 * @return PDR type.
	 */
	public static PDRType typeFromString(final String prefix)
	{
		if (prefix.equals(PDRType.ASPECT.toString()))
		{
			return PDRType.ASPECT;
		}
		else if (prefix.equals(PDRType.PERSON.toString()))
		{
			return PDRType.PERSON;
		}
		else if (prefix.equals(PDRType.REFERENCE.toString()))
		{
			return PDRType.REFERENCE;
		}
		else if (prefix.equals(PDRType.USER.toString()))
		{
			return PDRType.USER;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Returns the URL linked to the provided PDR type as a <code>java.lang.String</code>.
	 * @return XML Schema URL of PDR type.
	 */
	public final String getSchemaURL()
	{
		switch (_type)
		{
		case ASPECT:
			return SCHEMA_URL_ASPECT;
		case PERSON:
			return SCHEMA_URL_PERSON;
		case USER:
			return SCHEMA_URL_USER;
		case REFERENCE:
			return SCHEMA_URL_REFERENCE;
		default:
			return null;
		}
	}

	/**
	 * At risk of getting a <code>null</code> reference, this method can
	 * be used to create an instance of this class when the validity of
	 * the given <code>idString</code> is secured during runtime by other
	 * mechanims.
	 * @param idString PDR ID as string.
	 * @return <code>null</code> or an instance of this class.
	 */
	public static Identifier create(final String idString)
	{
		Identifier id = null;
		try
		{
			id =  new Identifier(idString);
		}
		catch (ValidationException e)
		{
			/* unhandled, see javadoc comment */
		}
		return id;
	}

	/**
	 * Converts this PDR identifier to a PaDRe identifier.
	 * @return PaDRe identifer.
	 * @throws ValidationException Invalid PaDRe identifier.
	 */
	public final PadreID toPadreID() throws ValidationException
	{
		StringBuilder sb = new StringBuilder();
		try
		{
			sb.append(_type.toString());
			sb.append('.');
			sb.append(Identifier.intToHexString(_instance, 3));
			sb.append('.');
			sb.append(Identifier.intToHexString(_project, 3));
			sb.append('.');
			sb.append(Identifier.intToHexString(_object, 9));
			return new PadreID(sb.toString());
		}
		catch (PadreIDException cause)
		{
			throw new ValidationException("tried to create invalid padre id: " + sb.toString());
		}
	}

	/**
	 * Converts the provided value to a hex string of the given length. If necessary,
	 * the string is filled up with zeros.
	 * @param value Value to convert.
	 * @param length Length of the resulting string.
	 * @return Hex string of given length.
	 */
	public static String intToHexString(final int value, final int length)
	{
		if (length <= 0)
		{
			throw new IllegalArgumentException("length has to be greater than 0 (was " + Integer.toString(length) + ")");
		}

		StringBuilder sb = new StringBuilder(Integer.toHexString(value));
		if (sb.length() > length)
		{
			throw new IllegalArgumentException("resulting string (" + sb.toString() + ") is longer than given length (" + Integer.toString(length) + ")");
		}
		else
		{
			while (sb.length() < length)
			{
				sb.insert(0, '0');
			}
			return sb.toString();
		}
	}
}
