/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.utils;

import java.text.ParseException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.apache.solr.common.SolrDocumentList;
import org.bbaw.pdr.padre.PadreCreateProject;
import org.bbaw.pdr.padre.PadreGetIDRanges;
import org.bbaw.pdr.padre.PadreGetInstance;
import org.bbaw.pdr.padre.PadreGetModified;
import org.bbaw.pdr.padre.PadreGetProjects;
import org.bbaw.pdr.padre.PadreGetTime;
import org.bbaw.pdr.padre.PadreSingleDelete;
import org.bbaw.pdr.padre.PadreSingleReadXml;
import org.bbaw.pdr.padre.PadreSpanReadXml;
import org.bbaw.pdr.padre.PadreWriteXml;
import org.bbaw.pdr.solr.SolrAddCollection;
import org.bbaw.pdr.solr.SolrCommit;
import org.bbaw.pdr.solr.SolrDelete;
import org.bbaw.pdr.solr.SolrOptimize;
import org.bbaw.pdr.solr.SolrPlainQuery;

/**
 * Encapsulates commonly useful methods for access to the repository
 * data. Protects some actions by the implementaion of a locking
 * mechanism which is available for all local methods, but not necessarily used.
 * @author Fabian Körner
 */
public final class Repository
{

	private static final Lock LOCK_STORAGE = new ReentrantLock(true);
	private static final Lock LOCK_INDEX = new ReentrantLock(true);
	private static final int LOCK_TIMEOUT_VALUE = 120;
	private static final TimeUnit LOCK_TIMEOUT_UNIT = TimeUnit.SECONDS;

	private static final Pattern PATTERN_TEMPORARY_ID = Pattern.compile("pdr[APRU]o\\.\\d{3}\\.\\d{3}\\.1\\d{8}");

	private Repository()
	{
	}

	/**
	 * Purges objects of the provided range from the repository.
	 * @param type PDR type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound.
	 * @param upperBound Upper bound.
	 * @throws RepositoryException Could not delete provided range.
	 */
	public static void deleteIDRange(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound)
			throws RepositoryException
	{
		boolean hasLocks = Repository.tryLockStorageAndIndex();
		try
		{
			int current = lowerBound;
			Identifier identifier;

			while (current <= upperBound)
			{
				identifier = new Identifier(type, instance, project, current);

				try
				{
					String id = identifier.toString();
					new PadreSingleDelete(id).execute();
					new SolrDelete("id:" + id).execute();
				}
				finally
				{
					current++;
				}
			}
			new SolrCommit().execute();
			new SolrOptimize().execute();
		}
		finally
		{
			if (hasLocks)
			{
				Repository.unlockStorageAndIndex();
			}
		}
	}

	/**
	 * Returns all PDR objects of the given project which were modified since the
	 * provided timestamp.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param timestamp Timestamp.
	 * @return Modified objects.
	 * @throws RepositoryException Could not get modified objects.
	 */
	public static List<String> getModifiedObjects(final int instance, final int project, final String timestamp) throws RepositoryException
	{
		List<String> result = new Vector<String>();

		boolean hasLocks = Repository.tryLockStorageAndIndex();
		if (hasLocks)
		{
			try
			{
				List<PDRObject> pdrObjects = new PadreGetModified(project, timestamp).execute();
				for (PDRObject pdrObject : pdrObjects)
				{
					result.add(pdrObject.getXMLContent());
				}
			}
			finally
			{
				Repository.unlockStorageAndIndex();
			}
		}
		else
		{
			throw new RepositoryException("couldn't get lock");
		}
		return result;
	}

	/**
	 * Returns all PDR objects of the given type and ID range.
	 * @param type PDR type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound.
	 * @param upperBound Upper bound.
	 * @return PDR objects.
	 * @throws RepositoryException Could not get objects.
	 */
	public static List<String> getObjects(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound)
			throws RepositoryException
	{
		final List<String> result = new Vector<String>();
		/* vergebene IDs im betrachteten Bereich */
		List<IDRange> ranges = Repository.getOccupiedObjectIDRanges(type, instance, project, lowerBound, upperBound);
		Identifier minId;
		Identifier maxId;
		for (IDRange range : ranges)
		{
			minId = new Identifier(type, instance, project, range.getLowerBound());
			maxId = new Identifier(type, instance, project, range.getUpperBound());

			PadreSpanReadXml padreCommand = new PadreSpanReadXml(minId.toString(), maxId.toString());
			List<PDRObject> objects = padreCommand.execute();

			for (PDRObject object : objects)
			{
				result.add(object.getXMLContent());
			}
		}
		return result;
	}

	/**
	 * Returns a list of occupied ID ranges in the repository.
	 * @param type PDR type.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param lowerBound Lower bound.
	 * @param upperBound Upper bound.
	 * @return Occupied ID ranges.
	 * @throws RepositoryException Could not get ranges.
	 */
	public static List<IDRange> getOccupiedObjectIDRanges(final PDRType type, final int instance, final int project, final int lowerBound, final int upperBound)
			throws RepositoryException
	{
		List<IDRange> ranges = new Vector<IDRange>();
		boolean hasLock = Repository.trylockStorage();
		if (hasLock)
		{
			try
			{
				ranges = new PadreGetIDRanges(type, instance, project, lowerBound, upperBound).execute();
			}
			finally
			{
				Repository.unlockStorage();
			}
		}
		else
		{
			throw new RepositoryException("couldn't get storage lock");
		}
		return ranges;
	}

	/**
	 * Returns a map of existing project IDs and the current number of their
	 * objects.
	 * @return Map of IDs and project sizes.
	 * @throws RepositoryException Could not get project sizes.
	 */
	public static Map<Integer, Integer> getProjectSizes() throws RepositoryException
	{
		final Map<Integer, Integer> result = new HashMap<Integer, Integer>();

		boolean hasLock = Repository.trylockStorage();
		if (hasLock)
		{
			int instance = -1;
			int project = -1;
			try
			{
				instance = Integer.parseInt(new PadreGetInstance().execute(), 16);
				List<String> projects = new PadreGetProjects().execute();

				int objectCount;
				for (String projectID : projects)
				{
					project = Integer.parseInt(projectID, 16);
					objectCount = Repository.countObjectsOfProject(instance, project);
					result.put(new Integer(project), new Integer(objectCount));
				}
			}
			finally
			{
				Repository.unlockStorage();
			}
		}
		else
		{
			throw new RepositoryException("couldn't get storage lock");
		}

		return result;
	}

	/**
	 * Returns current timestamp from PaDRe server.
	 * @return Timestamp from PaDRe server.
	 * @throws RepositoryException Communication failed.
	 */
	public static String getTime() throws RepositoryException
	{
		String result = null;
		result = new PadreGetTime().execute();
		result = result.trim();
		return result;
	}

	/**
	 * Returns the PDR user ID of an user identified by username and project number.
	 * @param username Username which is linked to the returned PDR user ID.
	 * @param project Number of user's project.
	 * @return PDR user ID.
	 * @throws RepositoryException Communication failed.
	 */
	public static Identifier getUserID(final String username, final int project) throws RepositoryException
	{
		Identifier result = null;
		boolean hasLock = Repository.tryLockIndex();
		try
		{
			String query = "id.type: pdrUo AND user.username:" + username + " AND id.project:" + Integer.toString(project);
			SolrQuery solrQuery = new SolrQuery(query);
			solrQuery.setRows(2);
			QueryResponse response = new SolrPlainQuery(solrQuery).execute();
			SolrDocumentList responseDocuments = response.getResults();
			if (1L == responseDocuments.getNumFound())
			{
				result = new Identifier((String) responseDocuments.get(0).getFieldValue("id"));
			}
			else if (0L == responseDocuments.getNumFound())
			{
				throw new RepositoryException("username " + username + " not found in project " + Integer.toString(project));
			}
			else
			{
				throw new RepositoryException("ambigous username " + username + " in project " + Integer.toString(project));
			}
		}
		finally
		{
			if (hasLock)
			{
				Repository.unlockIndex();
			}
		}
		return result;
	}

	/**
	 * Ingests provided new PDR objects (with temporary IDs) into the repository and
	 * returns a map from temporary to persistent IDs.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param objects PDR Objects to be inserted into the repository.
	 * @return Map from temporary to persistent IDs.
	 * @throws RepositoryException Problems while interacting with repository.
	 */
	public static Map<Identifier, Identifier> ingestObjects(final int instance, final int project, final List<String> objects) 	throws RepositoryException
	{
		List<PDRObject> objectsList = Repository.createPDRObjects(objects);

		Repository.checkPDRType(objectsList);
		Repository.checkInstanceID(instance, objectsList);
		Repository.checkProjectID(project, objectsList);

		boolean hasLocks = Repository.tryLockStorageAndIndex();
		if (hasLocks)
		{
			try
			{
				Repository.createProjectStore(project);

				for (PDRObject object : objectsList)
				{
					Repository.createPersistentAndReplaceTemporaryID(object, objectsList);
				}

				Repository.checkRemainingTemporaryIDs(objectsList);

				Repository.writeObjects(objectsList);
			}
			finally
			{
				Repository.unlockStorageAndIndex();
			}
		}
		else
		{
			throw new RepositoryException("couldn't get lock for ingestion");
		}
		return Repository.getMapFromOriginalToModifiedIdentifiers(objectsList);
	}

	/**
	 * Tries to store modified PDR object in the repository. If conflicts occur, a list
	 * of all conflicting object is returned.
	 * @param instance Instance ID.
	 * @param project Project ID.
	 * @param objects Modified objects.
	 * @param forceOverwrite If set to <code>true</code> all objects will be overwritten, regardless of any conflict.
	 * @return Liste mit XML-Repräsentationen von gespeicherten Objekten, bei denen Konflikte auftraten.
	 * @throws RepositoryException Modification of provided objects failed.
	 */
	public static List<String> modifyObjects(final int instance, final int project, final List<String> objects, final boolean forceOverwrite)
			throws RepositoryException
	{
		List<String> conflictsList = new Vector<String>();
		List<PDRObject> objectsList = Repository.createPDRObjects(objects);

		Repository.checkPDRType(objectsList);
		Repository.checkInstanceID(instance, objectsList);
		Repository.checkProjectID(project, objectsList);
		Repository.checkRemainingTemporaryIDs(objectsList);

		boolean hasLocks = Repository.tryLockStorageAndIndex();
		if (hasLocks)
		{
			try
			{
				if (forceOverwrite)
				{
					Repository.removeIndexDocuments(objectsList);
					Repository.writeObjects(objectsList);
				}
				else
				{
					conflictsList.addAll(Repository.tryToModifyObjects(objectsList));
				}
			}
			finally
			{
				Repository.unlockStorageAndIndex();
			}
		}
		else
		{
			throw new RepositoryException("couldn't get lock for modification");
		}
		return conflictsList;
	}

	/**
	 * Tries to lock the index component of the repository.
	 * @return Index is locked.
	 * @throws RepositoryException Could not lock index.
	 */
	public static synchronized boolean tryLockIndex() throws RepositoryException
	{
		try
		{
			return LOCK_INDEX.tryLock(LOCK_TIMEOUT_VALUE, LOCK_TIMEOUT_UNIT);
		}
		catch (InterruptedException cause)
		{
			throw new RepositoryException(cause);
		}
	}

	/**
	 * Tries to lock the storage component of the repository.
	 * @return Storage is locked.
	 * @throws RepositoryException Could not lock repository.
	 */
	public static synchronized boolean trylockStorage() throws RepositoryException
	{
		try
		{
			return LOCK_STORAGE.tryLock(LOCK_TIMEOUT_VALUE, LOCK_TIMEOUT_UNIT);
		}
		catch (InterruptedException cause)
		{
			throw new RepositoryException(cause);
		}
	}

	/**
	 * Tries to lock both storage, and index component of the repository.
	 * @return Index and storage are locked.
	 * @throws RepositoryException Could not lock index and storage.
	 */
	public static synchronized boolean tryLockStorageAndIndex() throws RepositoryException
	{
		return Repository.trylockStorage() && Repository.tryLockIndex();
	}

	/**
	 * Tries to unlock the index component of the repository.
	 */
	public static synchronized void unlockIndex()
	{
		LOCK_INDEX.unlock();
	}

	/**
	 * Tries to unlock the storage component of the repository.
	 */
	public static synchronized void unlockStorage()
	{
		LOCK_STORAGE.unlock();
	}

	/**
	 * Tries to unlock both index, and storage component of the repository.
	 */
	public static synchronized void unlockStorageAndIndex()
	{
		Repository.unlockStorage();
		Repository.unlockIndex();
	}

	private static void checkInstanceID(final int instance, final List<PDRObject> objects) throws RepositoryException
	{
		int currentInstance;
		for (PDRObject object : objects)
		{
			currentInstance = object.getInstance();
			if (currentInstance != instance)
			{
				throw new RepositoryException("invalid instance ID: " + Integer.toString(currentInstance) + " (expected: " + Integer.toString(instance) + ")");
			}
		}
	}

	private static void checkPDRType(final List<PDRObject> objects) throws RepositoryException
	{
		if (!objects.isEmpty())
		{
			PDRType firstType = objects.get(0).getType();
			PDRType currentType;
			for (PDRObject object : objects)
			{
				currentType = object.getType();
				if (firstType != currentType)
				{
					throw new RepositoryException("invalid type: " + currentType.toString() + " (expected: " + firstType.toString() + ")");
				}
			}
		}
	}

	private static void checkProjectID(final int project, final List<PDRObject> objects) throws RepositoryException
	{
		int currentProject;
		for (PDRObject object : objects)
		{
			currentProject = object.getProject();
			if (currentProject != project)
			{
				throw new RepositoryException("invalid project ID: " + Integer.toString(currentProject) + " (expected: " + Integer.toString(project) + ")");
			}
		}
	}

	/**
	 * Tests if any PDR object in the provided list contains one or more
	 * temporary ID. Such object must not be introduced into the repository.
	 * @param objects PDR objects.
	 * @throws RepositoryException Temporary ID has been found.
	 */
	public static void checkRemainingTemporaryIDs(final List<PDRObject> objects) throws RepositoryException
	{
		for (PDRObject object : objects)
		{
			Matcher matcher = PATTERN_TEMPORARY_ID.matcher(object.getXMLContent());
			if (matcher.find())
			{
				throw new RepositoryException("unresolved temporary ID(s) in PDR object: " + object.getXMLContent());
			}
		}
	}

	private static boolean conflictExists(final PDRObject modifiedObject, final PDRObject storedObject) throws RepositoryException
	{
		try
		{
			List<Date> timestampsModified = modifiedObject.getRevisionTimestamps();
			List<Date> timestampsStored = storedObject.getRevisionTimestamps();

			Date secondLastTimestampOfModifiedObject = timestampsModified.get(timestampsModified.size() - 2);
			Date lastTimestampOfStoredObject = timestampsStored.get(timestampsStored.size() - 1);

			return !(lastTimestampOfStoredObject.compareTo(secondLastTimestampOfModifiedObject) == 0);
		}
		catch (ParseException cause)
		{
			throw new RepositoryException(cause);
		}
	}

	private static int countObjectsOfProject(final int instanceID, final int projectID) throws RepositoryException
	{
		int result = 0;

		List<IDRange> idRanges;
		for (PDRType type : PDRType.values())
		{
			idRanges = new PadreGetIDRanges(type, instanceID, projectID, IDRange.MIN_OBJECT_ID, IDRange.MAX_OBJECT_ID).execute();
			for (IDRange idRange : idRanges)
			{
				result += idRange.count();
			}
		}

		return result;
	}

	private static List<PDRObject> createPDRObjects(final List<String> objects) throws ValidationException
	{
		List<PDRObject> result = new Vector<PDRObject>(objects.size());
		for (String object : objects)
		{
			object = object.trim();
			object = object.replaceAll("\n", "");
			object = object.replaceAll("xmlns:(aodl|podl|uodl|mods)", "xmlns");
			object = object.replaceAll("(aodl|podl|uodl|mods):", "");

			result.add(new PDRObject(object));
		}
		return result;
	}

	private static void createPersistentAndReplaceTemporaryID(final PDRObject object, final List<PDRObject> objects) throws RepositoryException
	{
		Identifier temporaryID = object.getOriginalIdentifier();
		Identifier persistentID = Repository.getNextPersistentIdentifier(object, objects);

		object.modifyIdentifier(persistentID);

		Repository.replaceIDsInObjects(objects, temporaryID.toString(), persistentID.toString());
	}

	private static void createProjectStore(final int project) throws RepositoryException
	{
		PadreGetProjects command = new PadreGetProjects();
		String projectString = Identifier.intToHexString(project, 3);
		List<String> projects = command.execute();
		if (!projects.contains(projectString))
		{
			new PadreCreateProject(project).execute();
		}
	}

	private static int getNextObjectID(final PDRType type, final int instance, final int project) throws RepositoryException
	{
		List<IDRange> ranges = Repository.getOccupiedObjectIDRanges(type, instance, project, IDRange.MIN_OBJECT_ID, IDRange.MAX_OBJECT_ID);
		if (ranges.isEmpty())
		{
			return IDRange.MIN_OBJECT_ID;
		}
		else
		{
			IDRange firstRange = ranges.get(0);
			if (firstRange.getLowerBound() > IDRange.MIN_OBJECT_ID)
			{
				return IDRange.MIN_OBJECT_ID;
			}
			else
			{
				return firstRange.getUpperBound() + 1;
			}
		}
	}

	private static int getNextObjectIDInList(final List<PDRObject> objects)
	{
		int result = 1;
		Identifier identifier;
		for (PDRObject object : objects)
		{
			identifier = object.getModifiedIdentifier();
			if (identifier.getObjectNumber() < 100000000)
			{
				result = Math.max(result, identifier.getObjectNumber() + 1);
			}
		}
		return result;
	}

	private static Identifier getNextPersistentIdentifier(final PDRObject original, final List<PDRObject> objects) throws RepositoryException
	{
		Identifier originalID = original.getOriginalIdentifier();
		int nextObjectIDInRepository = Repository.getNextObjectID(originalID.getType(), originalID.getInstanceNumber(), originalID.getProjectNumber());
		int nextObjectIDInList = Repository.getNextObjectIDInList(objects);
		int nextObjectID = Math.max(nextObjectIDInRepository, nextObjectIDInList);

		return new Identifier(originalID.getType(), originalID.getInstanceNumber(), originalID.getProjectNumber(), nextObjectID);
	}

	private static Map<Identifier, Identifier> getMapFromOriginalToModifiedIdentifiers(final List<PDRObject> objects)
	{
		Map<Identifier, Identifier> result = new HashMap<Identifier, Identifier>();
		for (PDRObject object : objects)
		{
			result.put(object.getOriginalIdentifier(), object.getModifiedIdentifier());
		}
		return result;
	}

	private static Pattern getPatternFromIDString(final String id)
	{
		StringBuilder sb = new StringBuilder(id);
		sb.insert(13, '\\');
		sb.insert(9, '\\');
		sb.insert(5, '\\');
		return Pattern.compile(sb.toString());
	}

	private static PDRObject getStoredObject(final Identifier identifier) throws RepositoryException
	{
		return new PadreSingleReadXml(identifier.toString()).execute();
	}

	private static void removeIndexDocuments(final List<PDRObject> objects) throws RepositoryException
	{
		for (PDRObject object : objects)
		{
			new SolrDelete("id:" + object.getModifiedIdentifier().toString()).execute();
		}
		new SolrCommit().execute();
	}

	private static void replaceIDsInObjects(final List<PDRObject> objects, final String original, final String replacement)
	{
		Pattern pattern = Repository.getPatternFromIDString(original);
		Matcher matcher;
		for (PDRObject object : objects)
		{
			matcher = pattern.matcher(object.getXMLContent());
			object.setXMLContent(matcher.replaceAll(replacement));
		}
	}

	private static List<String> tryToModifyObjects(final List<PDRObject> objects) throws RepositoryException
	{
		List<String> conflicts = new Vector<String>();
		PDRObject storedObject;
		for (PDRObject object : objects)
		{
			storedObject = Repository.getStoredObject(object.getModifiedIdentifier());
			if (Repository.conflictExists(object, storedObject))
			{
				conflicts.add(storedObject.getXMLContent());
			}
			else
			{
				List<PDRObject> objectToWrite = new Vector<PDRObject>(1);
				objectToWrite.add(object);
				Repository.removeIndexDocuments(objectToWrite);
				Repository.writeObjects(objectToWrite);
			}
		}
		return conflicts;
	}

	private static void writeObjects(final List<PDRObject> objects) throws RepositoryException
	{
		new PadreWriteXml(objects).execute();
		new SolrAddCollection(objects).execute();
		new SolrCommit().execute();
	}
}
